import numpy as np
from scipy.sparse import diags
from scipy.sparse import dok_matrix
from sklearn.metrics import pairwise_distances
from soynlp.utils import get_process_memory
from soynlp.vectorizer import sent_to_word_contexts_matrix

def pmi(x, min_pmi=0, alpha=0.0001, verbose=False):
    # convert x to probability matrix & marginal probability 
    px = (x.sum(axis=1) / x.sum()).reshape(-1)
    py = (x.sum(axis=0) / x.sum()).reshape(-1)
    pxy = x / x.sum()
    
    # transform px and py to diagonal matrix
    # using scipy.sparse.diags
    px_diag = diags(px.tolist()[0])
    py_diag = diags((py).tolist()[0])
    
    # pmi_alpha (x,y) = p(x,y) / ( p(x) x (p(y) + alpha) )
    px_diag.data[0] = np.asarray([0 if v == 0 else 1/v for v in px_diag.data[0]])
    py_diag.data[0] = np.asarray([0 if v == 0 else 1/(v + alpha) for v in py_diag.data[0]])
    
    exp_pmi = px_diag.dot(pxy).dot(py_diag)
    
    # PPMI using threshold
    min_exp_pmi = 1 if min_pmi == 0 else np.exp(min_pmi)

    # because exp_pmi is sparse matrix and type of exp_pmi.data is numpy.ndarray
    indices = np.where(exp_pmi.data > min_exp_pmi)[0]

    pmi_dok = dok_matrix(exp_pmi.shape)

    # prepare data (rows, cols, data)
    rows, cols = exp_pmi.nonzero()
    data = exp_pmi.data

    # enumerate function for printing status
    for _n_idx, idx in enumerate(indices):
        # print current status        
        if verbose and _n_idx % 10000 == 0:
            print('\rcomputing pmi {:.3} %  mem={} Gb    '.format(
                100 * _n_idx / indices.shape[0], '%.3f' % get_process_memory())
                  , flush=True, end='')
        # apply logarithm
        pmi_dok[rows[idx], cols[idx]] = np.log(data[idx])
    if verbose:
        print('\rcomputing pmi was done{}'.format(' '*30), flush=True)

    return pmi_dok

class PMI:
    def __init__(self, windows=3, min_tf=10, verbose=True,
                 tokenizer=lambda x:x.split(), min_pmi=0, alpha=0.0001):

        self.windows = windows
        self.min_tf = min_tf
        self.verbose = verbose
        self.tokenizer = tokenizer
        self.min_pmi = min_pmi
        self.alpha = alpha
    
    def train(self, sents):
        # construct word - context matrix
        self.x, self.idx2vocab = sent_to_word_contexts_matrix(
            sents, self.windows, self.min_tf, self.tokenizer, self.verbose)
        self.vocab2idx = {vocab:idx for idx, vocab in enumerate(self.idx2vocab)}

        # compute pmi
        self.pmi_ = pmi(self.x, min_pmi=self.min_pmi, alpha=self.alpha, verbose=self.verbose)

        return self

    def most_similar_words(self, query, topk=10, filter_func=lambda x:True):
        assert topk > 0

        if not (query in self.vocab2idx):
            return []

        query_idx = self.vocab2idx[query]
        dist = pairwise_distances(self.x[query_idx,:], self.x, metric='cosine')[0]
        similars = []
        for similar_idx in dist.argsort():
            if similar_idx == query_idx:
                continue

            if len(similars) >= topk:
                break

            similar_word = self.idx2vocab[similar_idx]
            if not filter_func(similar_word):
                continue

            similars.append((similar_word, 1-dist[similar_idx]))

        return similars

    def most_related_contexts(self, query, topk=10, filter_func=lambda x:True):
        assert topk > 0

        if not (query in self.vocab2idx):
            return []

        query_idx = self.vocab2idx[query]

        submatrix = self.pmi_[query_idx,:].tocsr()
        contexts = submatrix.nonzero()[1]
        pmi_i = submatrix.data
        most_relateds = [(idx, pmi_ij) for idx, pmi_ij in zip(contexts, pmi_i)]
        most_relateds = sorted(most_relateds, key=lambda x:-x[1])[:topk]
        most_relateds = [(self.idx2vocab[idx], pmi_ij) for idx, pmi_ij in most_relateds]

        return most_relateds


# -*- encoding:utf8 -*-

tagset = {
  "Noun": "명사",
  "Pronoun": "대명사",
  "Numeral": "수사",
  "Verb": "동사",
  "Adjective": "형용사",
  "Determiner": "관형사",
  "Adverb": "부사",
  "Exclamation": "감탄사",
  "Josa": "조사",
  "Symbol": "기호"
}


# -*- encoding:utf8 -*-

import sys
if sys.version_info <= (2,7):
    reload(sys)
    sys.setdefaultencoding('utf-8')
import json
import os

class Dictionary:
    def __init__(self, pos_dict):
        if isinstance(pos_dict, dict):
            for key in pos_dict:
                if sys.version_info.major == 2:
                    pos_dict[key] = set(map(unicode, pos_dict[key]))
                else:
                    pos_dict[key] = set(pos_dict[key])
            self.pos_dict = pos_dict
            self.max_length = self._check_max_length(self.pos_dict)
        elif isinstance(pos_dict, str):
            if os.path.exists(pos_dict):
                self.load(pos_dict)
            else:
                raise ValueError('dictionary file does not exist')

    def _check_max_length(self, pos_dict):
        return max((len(word) for words in pos_dict.values() for word in words))

    def get_pos(self, word):
        tags = []
        for pos, words in self.pos_dict.items():
            if word in words:
                tags.append(pos)
        return tags

    def word_is_tag(self, word, tag):
        return word in self.pos_dict.get(tag, {})

    def add_words(self, tag, words, force=False):
        words = self._type_check(words)

        if not force and not (tag in self.pos_dict):
            message = 'Check your tag or use add_words(tag, words, force=True)'.format(tag)
            raise ValueError(message)

        max_length = max((len(word) for word in words))
        if self.max_length < max_length:
            self.max_length = max_length

        if not (tag in self.pos_dict):
            dictionary = words
        else:
            dictionary = self.pos_dict.get(tag, {})
            dictionary.update(words)
        self.pos_dict[tag] = dictionary

    def remove_words(self, tag, words=None):
        if not (tag in self.pos_dict):
            raise ValueError('tag {} does not exist'.format(tag))

        if words == None:
            self.pos_dict.pop(tag)
            return

        words = self._type_check(words)
        dictionary = self.pos_dict[tag]
        dictionary -= words

    def _type_check(self, words):
        if isinstance(words, str):
            words = set(words.split())
        return words

    def load(self, filename):
        if sys.version_info.major == 2:
            with open(filename) as fp:
                params = json.load(unicode(fp))
                self.max_length = params['max_length']
                self.pos_dict = {tag:set(words)
                    for tag, words in params['pos_dict'].items()}
        else:
            with open(filename, encoding= "utf-8") as fp:
                params = json.load(fp)
                self.max_length = params['max_length']
                self.pos_dict = {tag:set(words)
                    for tag, words in params['pos_dict'].items()}

    def save(self, filename):
        if sys.version_info.major == 2:
            with open(filename, 'w') as fp:
                params = {
                    'max_length':self.max_length,
                    'pos_dict': {pos:list(words)
                        for pos, words in self.pos_dict.items()}
                }
                json.dump(params, fp, ensure_ascii=False, indent=2)
        else:
            with open(filename, 'w', encoding= "utf-8") as fp:
                params = {
                    'max_length':self.max_length,
                    'pos_dict': {pos:list(words)
                        for pos, words in self.pos_dict.items()}
                }
                json.dump(params, fp, ensure_ascii=False, indent=2)


# -*- encoding:utf8 -*-

import re
from soynlp.hangle import decompose, compose

repeatchars_patterns = [
    re.compile('(\w\w\w\w)\\1{3,}'),
    re.compile('(\w\w\w)\\1{3,}'),
    re.compile('(\w\w)\\1{3,}'),
    re.compile('(\w)\\1{3,}')
]

def normalize(sentence, num_repeat=2):
    tokens = sentence.split()
    return ' '.join(_normalize_korean_token(token, num_repeat) for token in tokens)

def _normalize_korean_token(token, num_repeat=2):
    token = _normalize_emoji(token)
    token = _remove_repeat(token, num_repeat)
    return token

def _remove_repeat(token, num_repeat=2):
    if num_repeat > 0:
        for pattern in repeatchars_patterns:
            token = pattern.sub('\\1' * num_repeat, token)
    return token

def _normalize_emoji(token):
    if len(token) <= 1:
        return token
    token_ = []
    decomposeds = [decompose(c) for c in token]
    for char, cd, nd in zip(token, decomposeds, decomposeds[1:]):
        if cd == None or nd == None:
            token_.append(char)
            continue
        # 앜ㅋㅋㅋㅋ -> 아ㅋㅋㅋㅋㅋ
        if (nd[1] == ' ') and (cd[2] == nd[0]):
            token_.append(compose(cd[0], cd[1], ' ') + nd[0])
        # ㅋ쿠ㅜㅜ -> ㅋㅋㅜㅜㅜ
        elif (cd[2] == ' ') and (nd[0] == ' ') and (cd[1] == nd[1]):
            token_.append((cd[0] + cd[1]) if cd [0] != ' ' else cd[1])
        else:
            token_.append(char)
    return ''.join(token_) + token[-1]


# -*- encoding:utf8 -*-

import os
import psutil
import sys
from collections import defaultdict

installpath = os.path.sep.join(
    os.path.dirname(os.path.realpath(__file__)).split(os.path.sep)[:-1])

def get_available_memory():
    """It returns remained memory as percentage"""

    mem = psutil.virtual_memory()
    return 100 * mem.available / (mem.total)

def get_process_memory():
    """It returns the memory usage of current process"""
    
    process = psutil.Process(os.getpid())
    return process.memory_info().rss / (1024 ** 3)

def check_dirs(filepath):
    dirname = os.path.dirname(filepath)
    if dirname and dirname == '.' and not os.path.exists(dirname):
        os.makedirs(dirname)
        print('created {}'.format(dirname))

def sort_by_alphabet(filepath):
    if sys.version.split('.')[0] == '2':
        with open(filepath) as f:
            docs = [doc.strip() for doc in f]
            docs = [doc for doc in docs if doc]
    else:
        with open(filepath, encoding= "utf-8") as f:
            docs = [doc.strip() for doc in f]
            docs = [doc for doc in docs if doc]
    if sys.version.split('.')[0] == '2':
        with open(filepath, 'w') as f:
            for doc in sorted(docs):
                f.write('{}\n'.format(doc))
    else:
        with open(filepath, 'w', encoding= "utf-8") as f:
            for doc in sorted(docs):
                f.write('{}\n'.format(doc))

class DoublespaceLineCorpus:    
    def __init__(self, corpus_fname, num_doc = -1, num_sent = -1, iter_sent = False, skip_header = 0):
        self.corpus_fname = corpus_fname
        self.num_doc = 0
        self.num_sent = 0
        self.iter_sent = iter_sent
        self.skip_header = skip_header
        if (num_doc > 0) or (num_sent > 0):
            self.num_doc, self.num_sent = self._check_length(num_doc, num_sent)

    def _check_length(self, num_doc, num_sent):
        num_sent_ = 0
        try:
            try:
                # python version check
                if sys.version.split('.')[0] == '2':
                    f = open(self.corpus_fname)
                else:
                    f = open(self.corpus_fname, encoding= "utf-8")

                # skip headers
                for _ in range(self.skip_header):
                    next(f)
                
                # check length
                for doc_idx, doc in enumerate(f):
                    if (num_doc > 0) and (doc_idx >= num_doc):
                        return doc_idx, num_sent_
                    sents = doc.split('  ')
                    sents = [sent for sent in sents if sent.strip()]
                    num_sent_ += len(sents)
                    if (num_sent > 0) and (num_sent_ > num_sent):
                        return doc_idx+1, min(num_sent, num_sent_)

            finally:
                f.close()
            return doc_idx+1, num_sent_

        except Exception as e:
            print(e)
            return -1, -1
    
    def __iter__(self):
        try:
            try:
                if sys.version.split('.')[0] == '2':
                    f = open(self.corpus_fname)
                else:
                    f = open(self.corpus_fname, encoding='utf-8')
                
                # skip headers
                for _ in range(self.skip_header):
                    next(f)
                    
                # iteration
                num_sent, stop = 0, False
                for doc_idx, doc in enumerate(f):
                    if stop:
                        break

                    # yield doc
                    if not self.iter_sent:
                        yield doc
                        if (self.num_doc > 0) and ((doc_idx + 1) >= self.num_doc):
                            stop = True
                        continue

                    # yield sents
                    for sent in doc.split('  '):
                        if (self.num_sent > 0) and (num_sent >= self.num_sent):
                            stop = True
                            break
                        sent = sent.strip()
                        if sent:
                            yield sent
                            num_sent += 1
            finally:
                f.close()

        except Exception as e:
            print(e)

    def __len__(self):
        if self.num_doc == 0:
            self.num_doc, self.num_sent = self._check_length(-1, -1)
        return self.num_sent if self.iter_sent else self.num_doc

class EojeolCounter:
    def __init__(self, sents=None, min_count=1,
        max_length=15, filtering_checkpoint=0, verbose=False):

        self.min_count = min_count
        self.max_length = max_length
        self.filtering_checkpoint = filtering_checkpoint
        self.verbose = verbose
        self._coverage = 0.0
        if sents:
            self._counter = self._counting_from_sents(sents)
        else:
            self._counter = {}
        self._count_sum = sum(self._counter.values())

    def __getitem__(self, eojeol):
        return self._counter.get(eojeol, 0)

    def _counting_from_sents(self, sents):
        _counter = {}
        for i_sent, sent in enumerate(sents):
            # filtering during eojeol counting
            if (self.min_count > 1 and
                self.filtering_checkpoint > 0 and
                i_sent > 0 and
                i_sent % self.filtering_checkpoint == 0):
                _counter = {k:v for k,v in _counter.items()
                            if v >= self.min_count}
            # add eojeol count
            for eojeol in sent.split():
                if (not eojeol) or (len(eojeol) <= 1) or (len(eojeol) > self.max_length):
                    continue
                _counter[eojeol] = _counter.get(eojeol, 0) + 1
            # print status
            if self.verbose and i_sent % 100000 == 99999:
                print('\r[EojeolCounter] n eojeol = {} from {} sents. mem={} Gb{}'.format(
                    len(_counter), i_sent + 1, '%.3f'%get_process_memory(), ' '*20), flush=True, end='')
        # final filtering
        _counter = {k:v for k,v in _counter.items() if v >= self.min_count}
        if self.verbose:
            print('\r[EojeolCounter] n eojeol = {} from {} sents. mem={} Gb{}'.format(
                len(_counter), i_sent + 1, '%.3f'%get_process_memory(), ' '*20), flush=True)
        return _counter

    @property
    def coverage(self):
        return self._coverage

    @coverage.setter
    def coverage(self, value):
        if not (0 <= value <= 1):
            raise ValueError('coverage should be in [0, 1]')
        self._coverage = value

    @property
    def num_of_unique_uncovered_eojeols(self):
        return len(self._counter)

    @property
    def num_of_uncovered_eojeols(self):
        return sum(self._counter.values())

    def get_uncovered_eojeols(self, min_count=0):
        return {k:v for k,v in self._counter.items() if v >= min_count}

    def remove_covered_eojeols(self, eojeols):
        self._counter = {k:v for k,v in self._counter.items() if not (k in eojeols)}
        self.coverage = 1 - self.num_of_uncovered_eojeols / self._count_sum

    def get_eojeol_count(self, eojeol):
        return self._counter.get(eojeol, 0)

    def items(self):
        return self._counter.items()

    def to_lrgraph(self, l_max_length=10, r_max_length=9):
        return self._to_lrgraph(self._counter, l_max_length, r_max_length)

    def _to_lrgraph(self, counter, l_max_length=10, r_max_length=9):
        _lrgraph = defaultdict(lambda: defaultdict(int))
        for eojeol, count in counter.items():
            for e in range(1, min(l_max_length, len(eojeol)) + 1):
                l, r = eojeol[:e], eojeol[e:]
                if len(r) > r_max_length:
                    continue
                _lrgraph[l][r] += count
        _lrgraph = {l:dict(rdict) for l, rdict in _lrgraph.items()}
        lrgraph = LRGraph(lrgraph=_lrgraph,
            l_max_length=l_max_length, r_max_length=r_max_length)
        return lrgraph

    def save(self, path):
        dirname = os.path.dirname(path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(path, 'w', encoding='utf-8') as f:
            for eojeol, count in sorted(self._counter.items(), key=lambda x:(-x[1], x[0])):
                f.write('{} {}\n'.format(eojeol, count))

    def load(self, path):
        self._coverage = 0.0
        self._counter = {}
        with open(path, encoding='utf-8') as f:
            for line in f:
                word, count = line.split()
                self._counter[word] = int(count)
        self._count_sum = sum(self._counter.values())

class LRGraph:

    def __init__(self, lrgraph=None, sents=None, l_max_length=10, r_max_length=9):

        assert l_max_length > 1 and type(l_max_length) == int
        assert r_max_length > 0 and type(r_max_length) == int

        self.l_max_length = l_max_length
        self.r_max_length = r_max_length

        if sents:
            if lrgraph:
                raise ValueError(
                    'Inserted lrgraph will be ignored. Insert only one (lrgraph, sents)')
            lrgraph = self._construct_graph(sents)
        if lrgraph:
            self._lr, self._rl = self._check_lrgraph(lrgraph)
        else:
            self._lr, self._rl = {}, {}

        self._lr_origin = {l:{r:c for r,c in rdict.items()}
                           for l,rdict in self._lr.items()}

    def _construct_graph(self, sents):
        lrgraph = defaultdict(lambda: defaultdict(int))
        for sent in sents:
            for word in sent.split():
                word = word.strip()
                for e in range(1, min(len(word), self.l_max_length) + 1):
                    l, r = word[:e], word[e:]
                    if len(r) > self.r_max_length:
                        continue
                    lrgraph[l][r] += e
        lrgraph = {l:dict(rdict) for l,rdict in lrgraph.items()}
        return lrgraph

    def _check_lrgraph(self, lrgraph):
        if type(lrgraph) is not dict:
            try:
                lrgraph = dict(lrgraph)
            except:
                raise ValueError('lrgraph type should be dict of dict, not {}'.format(
                    type(lrgraph)))
        nested_dict_type = type(list(lrgraph.values())[0])
        if nested_dict_type == defaultdict:
            lrgraph = {l:dict(rdict) for l,rdict in lrgraph.items()}
        elif not (nested_dict_type == dict):
            raise ValueError('nested value type should be dict, not {}'.format(
                nested_dict_type))
        rlgraph = defaultdict(lambda: defaultdict(int))
        for l, rdict in lrgraph.items():
            for r, c in rdict.items():
                if not r:
                    continue
                rlgraph[r][l] += c
        rlgraph = {r:dict(ldict) for r, ldict in rlgraph.items()}
        return lrgraph, rlgraph

    def reset_lrgraph(self):
        if not self._lr_origin:
            return None

        self._lr, self._rl = self._check_lrgraph(
            {l:{r:c for r,c in rdict.items()}
             for l, rdict in self._lr_origin.items()}
        )

    def add_lr_pair(self, l, r, count=1):
        self._lr[l][r] += count
        if r:
            self._rl[r][l] += count

    def add_eojeol(self, eojeol, count=1):
        for i in range(1, len(eojeol) + 1):
            l, r = eojeol[:i], eojeol[i:]
            self.add_lr_pair(l, r, count)

    def remove_lr_pair(self, l, r, count=1):
        if l in self._lr:
            rdict = self._lr[l]
            if r in rdict:
                rdict[r] -= count
                if rdict[r] <= 0:
                    rdict.pop(r)
                    if len(rdict) <= 0:
                        self._lr.pop(l)
        if r in self._rl:
            ldict = self._rl[r]
            if l in ldict:
                ldict[l] -= count
                if ldict[l] <= 0:
                    ldict.pop(l)
                    if len(ldict) <= 0:
                        self._rl.pop(r)

    def remove_eojeol(self, eojeol, count=1):
        for i in range(1, len(eojeol) + 1):
            l, r = eojeol[:i], eojeol[i:]
            self.remove_lr_pair(l, r, count)

    def get_r(self, l, topk=10):
        rlist = sorted(self._lr.get(l, {}).items(), key=lambda x:-x[1])
        if topk > 0:
            rlist = rlist[:topk]
        return rlist

    def get_l(self, r, topk=10):
        llist = sorted(self._rl.get(r, {}).items(), key=lambda x:-x[1])
        if topk > 0:
            llist = llist[:topk]
        return llist

    def freeze(self):
        """Remove self._lr_origin. Be careful.
        When you excute freeze, you cannot reset_lrgraph anynore."""
        self._lr_origin = None

    def copy_compatified_lrgraph_origin(self):
        """It returns original LRGraph which has no self._lr_origin"""
        lr_graph = LRGraph(
            l_max_length = self.l_max_length,
            r_max_length = self.r_max_length)
        lr_graph._lr, lr_graph._rl = self._check_lrgraph(
            {l:{r:c for r,c in rdict.items()}
             for l, rdict in self._lr_origin.items()}
        )
        return lr_graph

    def to_EojeolCounter(self, reset_lrgraph=False):
        lr = self._lr_origin if reset_lrgraph else self._lr
        counter = {}
        for l, rdict in lr.items():
            for r, count in rdict.items():
                counter[l+r] = count
        eojeol_counter = EojeolCounter(None)
        eojeol_counter._counter = counter
        eojeol_counter._count_sum = sum(counter.values())
        return eojeol_counter

    def save(self, path):
        dirname = os.path.dirname(path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(path, 'w', encoding='utf-8') as f:
            for l, rdict in sorted(self._lr_origin.items()):
                for r, c in sorted(rdict.items()):
                    f.write('{} {} {}\n'.format(l, r, c))

    def load(self, path):
        self._lr_origin = {}
        with open(path, encoding='utf-8') as f:
            l = ''
            rdict = {}
            for line in f:
                sep = line.split()
                if not (sep[0] == l):
                    if rdict:
                        self._lr_origin[l] = rdict
                        rdict = {}
                l = sep[0]
                if len(sep) == 2:
                    rdict[''] = int(sep[-1])
                elif len(sep) == 3:
                    rdict[sep[1]] = int(sep[-1])
                else:
                    raise ValueError('Wrong lr-graph format: {}'.format(line))
            if rdict:
                self._lr_origin[l] = rdict
        self._lr, self._rl = self._check_lrgraph(
            {l:{r:c for r,c in rdict.items()}
             for l,rdict in self._lr_origin.items()})



# -*- encoding:utf8 -*-

from collections import defaultdict, namedtuple
import math
import sys
from soynlp.word import WordExtractor
from soynlp.utils import LRGraph

NounScore = namedtuple('NounScore', 'frequency score known_r_ratio')

class LRNounExtractor:

    def __init__(self, max_left_length=10, max_right_length=7,
            predictor_fnames=None, verbose=True, min_num_of_features=1):

        self.coefficient = {}
        self.verbose = verbose
        self.max_left_length = max_left_length
        self.max_right_length = max_right_length
        self.lrgraph = None
        self.words = None
        self._substring_counter = {}
        self.min_num_of_features = min_num_of_features

        if not predictor_fnames:
            import os
            directory = '/'.join(os.path.abspath(__file__).replace('\\', '/').split('/')[:-2])
            predictor_fnames = ['%s/trained_models/noun_predictor_sejong' % directory]
            if verbose:
                print('used default noun predictor; Sejong corpus predictor')

        for fname in predictor_fnames:
            if verbose:
                print('used %s' % fname.split('/')[-1])
            self._load_predictor(fname)
        if verbose:
            print('All %d r features was loaded' % len(self.coefficient))

    def _load_predictor(self, fname):
        try:
            if sys.version_info.major == 2:
                f = open(fname)
            else:
                f = open(fname, encoding='utf-8')
            try:
                for num_line, line in enumerate(f):
                    r, score = line.strip().split('\t')
                    score = float(score)
                    if r in self.coefficient:
                        self.coefficient[r] = max(self.coefficient[r], score)
                    else:
                        self.coefficient[r] = score
            except Exception as e:
                print('predictor parsing error line {} = {}'.format(num_line+1, line))
            finally:
                f.close()
        except Exception as e:
            print(e)

    def train_extract(self, sents, min_noun_score=0.5, min_noun_frequency=5,
            noun_candidates=None):

        self.train(sents, min_noun_frequency)
        return self.extract(min_noun_score, min_noun_frequency, noun_candidates)

    def train(self, sents, min_noun_frequency=5):
        wordset_l, wordset_r = self._scan_vocabulary(sents, min_noun_frequency)
        lrgraph = self._build_lrgraph(sents, wordset_l, wordset_r)
        self.lrgraph = LRGraph(lrgraph)
        self.words = wordset_l

    def _scan_vocabulary(self, sents, min_frequency=5):
        """
        Parameters
        ----------
            sents: list-like iterable object which has string
            
        It computes subtoken frequency first. 
        After then, it builds lr-graph with sub-tokens appeared at least min count
        """

        _ckpt = int(len(sents) / 40)

        wordset_l = defaultdict(lambda: 0)
        wordset_r = defaultdict(lambda: 0)

        for i, sent in enumerate(sents):
            for token in sent.split(' '):
                if not token:
                    continue
                token_len = len(token)
                for i in range(1, min(self.max_left_length, token_len)+1):
                    wordset_l[token[:i]] += 1
                for i in range(1, min(self.max_right_length, token_len)):
                    wordset_r[token[-i:]] += 1
            if self.verbose and (i % _ckpt == 0):
                args = ('#' * int(i/_ckpt), '-' * (40 - int(i/_ckpt)), 100.0 * i / len(sent), '%')
                sys.stdout.write('\rscanning: %s%s (%.3f %s)' % args)

        self._substring_counter = {w:f for w,f in wordset_l.items() if f >= min_frequency}
        wordset_l = set(self._substring_counter.keys())
        wordset_r = {w for w,f in wordset_r.items() if f >= min_frequency}

        if self.verbose:
            print('\rscanning completed')
            print('(L,R) has (%d, %d) tokens' % (len(wordset_l), len(wordset_r)))

        return wordset_l, wordset_r

    def _build_lrgraph(self, sents, wordset_l, wordset_r):
        _ckpt = int(len(sents) / 40)
        lrgraph = defaultdict(lambda: defaultdict(lambda: 0))

        for i, sent in enumerate(sents):
            for token in sent.split():
                if not token:
                    continue
                n = len(token)
                for i in range(1, min(self.max_left_length, n)+1):
                    l = token[:i]
                    r = token[i:]
                    if not (l in wordset_l):
                        continue
                    if (len(r) > 0) and not (r in wordset_r):
                        continue
                    lrgraph[l][r] += 1

            if self.verbose and (i % _ckpt == 0):
                args = ('#' * int(i/_ckpt), '-' * (40 - int(i/_ckpt)), 100.0 * i / len(sents), '%')
                sys.stdout.write('\rbuilding lr-graph: %s%s (%.3f %s)' % args)
        if self.verbose:
            sys.stdout.write('\rbuilding lr-graph completed')
        lrgraph = {l:{r:f for r,f in rdict.items()} for l,rdict in lrgraph.items()}
        return lrgraph

    def extract(self, min_noun_score=0.5, min_noun_frequency=5, noun_candidates=None):
        if not noun_candidates:
            noun_candidates = self.words

        # prediction
        nouns = {}
        for word in sorted(noun_candidates, key=lambda w:len(w)):
            if len(word) <= 1:
                continue

            score = self.predict(word, nouns)

            if score[0] < min_noun_score:
                continue
            nouns[word] = score

        # postprocessing
        nouns = self._postprocess(nouns, min_noun_score, min_noun_frequency)

        # summary information as NounScore
        nouns_ = self._to_NounScore(nouns)

        return nouns_

    def _get_r_features(self, word):
        features = self.lrgraph.get_r(word, -1)
        # remove empty str r only in features
        features = [feature for feature in features if feature[0]]
        return features

    def _get_subword_score(self, word, min_noun_score, nouns):
        subword_scores = {}
        for e in range(1, len(word)):
            subword = word[:e]
            suffix = word[e:]
            # Add word if compound
            if (subword in nouns) and (suffix in nouns):
                score1 = nouns[subword]
                score2 = nouns[suffix]
                subword_scores[subword] = max(score1, score2)
            elif (subword in nouns) and (self.coefficient.get(suffix,0.0) > min_noun_score):
                subword_scores[subword] = (self.coefficient.get(suffix,0.0), 0)
        if not subword_scores:
            return (0.0, 0)
        return sorted(subword_scores.items(), key=lambda x:-x[1][0])[0][1]

    def is_noun(self, word, min_noun_score=0.5):
        return self.predict(word)[0] >= min_noun_score

    def predict(self, word, min_noun_score=0.5, nouns=None):
        """Returns (noun_score, known_r_ratio)
        """
        features = self._get_r_features(word)

        # (감사합니다 + 만) 처럼 뒤에 등장하는 R 의 종류가 한가지 뿐이면 제대로 된 판단이 되지 않음
        if len(features) > self.min_num_of_features:
            score = self._predict(features, word)
        else:
            if nouns is None:
                nouns = {}
            score = self._get_subword_score(word, min_noun_score, nouns)

        return score

    def _predict(self, features, word):

        def exist_longer_r_feature(word, r):
            for e in range(len(word)-1, -1, -1):
                suffix = word[e:] + r
                if suffix in self.coefficient: 
                    return True
            return False

        """Parameters
        ----------
            features: dict
                예시: {을: 35, 는: 22, ...}
        """

        score = 0
        norm = 0
        unknown = 0

        for r, freq in features:
            if r in self.coefficient:
                if not exist_longer_r_feature(word, r):
                    score += freq * self.coefficient[r]
                    norm += freq
            else:
                unknown += freq

        return (0 if norm == 0 else score / norm,
                0 if (norm + unknown == 0) else norm / (norm + unknown))

    def _postprocess(self, nouns, min_noun_score, min_noun_frequency):

        def is_Noun_Josa(l, r):
            return (l in nouns) and (self.coefficient.get(r, 0.0) > min_noun_score)

        def cohesion(word):
            base = self._substring_counter.get(word[0], 0)
            n = len(word)
            if not base or n <= 1:
                return 0
            return math.pow(self._substring_counter.get(word, 0) / base, 1/(n-1))

        def longer_has_larger_cohesion(word):
            return cohesion(word) >= cohesion(word[:-1])

        removals = set()
        for word in nouns:
            if word[-1] == '.' or word[-1] == ',':
                removals.add(word)
                continue
            n = len(word)
            if n <= 2 or longer_has_larger_cohesion(word):
                continue
            for e in range(2, len(word)):
                l = word[:e]
                r = word[e:]
                if is_Noun_Josa(l, r):
                    removals.add(word)
                    break
        nouns_ = {word:score for word, score in nouns.items() if (word in removals) == False}
        return nouns_

    def _to_NounScore(self, nouns):
        noun_frequencies = {}
        for word in sorted(nouns, key=lambda x:-len(x)):
            r_count = self.lrgraph.get_r(word, -1)
            noun_frequencies[word] = sum(c for w, c in r_count)
            for r, count in r_count:
                self.lrgraph.remove_eojeol(word+r, count)
        self.lrgraph.reset_lrgraph()

        nouns_ = {}
        for word, score in nouns.items():
            nouns_[word] = NounScore(noun_frequencies[word], score[0], score[1])

        return nouns_



from soynlp.noun import LRNounExtractor_v2
from soynlp.predicator import PredicatorExtractor
from soynlp.utils import LRGraph

class POSExtractor:

    def __init__(self, l_max_length=10, r_max_length=8, verbose=True, logpath=None):
        self.l_max_length = l_max_length
        self.r_max_length = r_max_length
        self.verbose = verbose
        self.logpath = logpath

    def _print(self, message, replace=False, newline=True):
        header = '[POS Extractor]'
        if replace:
            print('\r{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)
        else:
            print('{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)

    def extract(self, sentences):

        self._num_of_eojeols = 0
        self._num_of_covered_eojeols = 0

        nouns = self._extract_nouns(sentences)
        predicators = self._extract_predicators(self._lrgraph, nouns)

        del self._lrgraph

        if self.verbose:
            message = '{} nouns, {} predicators were extracted'.format(
                len(nouns), len(predicators))
            self._print(message, replace=True, newline=True)

        return nouns, predicators

    def _extract_nouns(self, sentences):

        noun_extractor = LRNounExtractor_v2(
            l_max_length = self.l_max_length,
            r_max_length = self.r_max_length,
            min_eojeol_count = 2,
            min_num_of_features = 2,
            max_count_when_noun_is_eojeol = 15,
            extract_compound = False,
            logpath = self.logpath,
            extract_pos_feature = True,
            verbose = self.verbose
        )

        noun_extractor.train(sentences)

        nouns = noun_extractor.extract(
            reset_lrgraph = False,
            min_count = 10,
            minimum_noun_score = 0.4,
        )

        self._lrgraph = LRGraph(
            {l:{r:v for r,v in rdict.items()}
             for l, rdict in noun_extractor.lrgraph._lr.items()}
        )
        self._num_of_eojeols = noun_extractor._num_of_eojeols
        self._num_of_covered_eojeols = noun_extractor._num_of_covered_eojeols

        self.noun_extractor = noun_extractor

        if self.verbose:
            message = 'noun extraction was done. {} % eojeols are covered'.format(
                '%.2f' % (100 * self._num_of_covered_eojeols / self._num_of_eojeols))
            self._print(message, replace=True, newline=True)

        return nouns

    def _extract_predicators(self, sentences_or_lrgraph, nouns=None):

        predicator_extractor = PredicatorExtractor(
            nouns,
            extract_eomi=True,
            extract_stem=True,
            verbose = self.verbose
        )

        predicator_extractor.train(sentences_or_lrgraph, min_eojeol_count=2)

        predicators = predicator_extractor.extract(
            candidates=None, min_count=10, reset_lrgraph=True,
            # Eomi extractor
            min_num_of_features=5, min_eomi_score=0.3, min_eomi_frequency=1,
            # Stem extractor
            min_num_of_unique_R_char=10, min_entropy_of_R_char=0.5,
            min_entropy_of_R=1.5, min_stem_score=0.7, min_stem_frequency=100
        )

        self._num_of_covered_eojeols += predicator_extractor._num_of_covered_eojeols

        self.predicator_extractor = predicator_extractor

        if self.verbose:
            message = 'predicator extraction was done. {} % eojeols are covered (cum)'.format(
                '%.2f' % (100 * self._num_of_covered_eojeols / self._num_of_eojeols))
            self._print(message, replace=True, newline=True)

        return predicators


# -*- encoding:utf8 -*-

from collections import namedtuple
NewsNounScore = namedtuple('NewsNounScore', 'score frequency feature_proportion eojeol_proportion n_positive_feature unique_positive_feature_proportion')
import sys

class NewsNounExtractor:
    
    def __init__(self, max_left_length=10, max_right_length=7,
            predictor_fnames=None, verbose=True, base_noun_dictionary=None):

        self.max_left_length = max_left_length
        self.max_right_length = max_right_length
        self.verbose = verbose
        self.r_scores = {}
        self.noun_dictionary = noun_dictionary if base_noun_dictionary else {}

        import os
        directory = '/'.join(os.path.abspath(__file__).replace('\\', '/').split('/')[:-2])

        if not predictor_fnames:
            predictor_fnames = ['%s/trained_models/noun_predictor_sejong' % directory]
            if verbose:
                print('used default noun predictor; Sejong corpus based logistic predictor')

        for fname in predictor_fnames:
            self._load_predictor(fname)
        self.josa_dictionary = {r for r, s in self.r_scores.items() if s > 0.1}
        self.josa_dictionary.update({'는'})

        print(directory)

        self._vdictionary = set()
        self._vdictionary.update(self._load_dictionary('%s/pos/dictionary/sejong/Verb.txt' % directory))
        self._vdictionary.update(self._load_dictionary('%s/pos/dictionary/sejong/Adjective.txt' % directory))

    def _load_predictor(self, fname):
        try:
            try:
                if sys.version_info.major == 2:
                    f = open(fname)
                else:
                    f = open(fname, encoding='utf-8')
                
                for num_line, line in enumerate(f):
                    r, score = line.strip().split('\t')
                    score = float(score)
                    if r in self.r_scores:
                        self.r_scores[r] = max(self.r_scores[r], score)
                    else:
                        self.r_scores[r] = score
            finally:
                f.close()
        except Exception as e:
            print(e)

    def _load_dictionary(self, fname):
        try:
            try:
                if sys.version_info.major == 2:
                    f = open(fname)
                else:
                    f = open(fname, encoding='utf-8')
                words = {word.strip().split('\t')[0] for word in f}
            finally:
                f.close()                    
            return words
        except Exception as e:
            print(e)
            return set()

    def train_extract(self, sents, min_frequency=3, min_noun_score=0.4,
            noun_candidates=None, min_feature_proportion=0.6):

        self.train(sents)
        return self.extract(min_noun_score, min_frequency,
            noun_candidates, min_feature_proportion)

    def train(self, sents):
        if self.verbose:
            print('scan vocabulary ... ')

        self.lrgraph, self.rlgraph, self.eojeols = self._build_graph(sents)
        self.lcount = {k:sum(d.values()) for k,d in self.lrgraph.items()}
        self.rcount = {k:sum(d.values()) for k,d in self.rlgraph.items()}

        if self.verbose:
            print('done (Lset, Rset, Eojeol) = ({}, {}, {})'.format(
                len(self.lcount), len(self.rcount), len(self.eojeols)))

    def _build_graph(self, sents):
        from collections import defaultdict
        from collections import Counter
        dictdictize = lambda dd: {k:dict(d) for k,d in dd.items()}

        max_eojeol_length = self.max_left_length + self.max_right_length
        eojeols = Counter(
            (eojeol for sent in sents for eojeol in sent.split() 
             if len(eojeol) <= max_eojeol_length)
        )

        lrgraph = defaultdict(lambda: defaultdict(lambda: 0))
        rlgraph = defaultdict(lambda: defaultdict(lambda: 0))
        for eojeol, count in eojeols.items():
            n = len(eojeol)
            for e in range(1, min(self.max_left_length, n) + 1):
                if (n - e) > self.max_right_length:
                    continue
                (l, r) = (eojeol[:e], eojeol[e:])
                lrgraph[l][r] += count
                if r:
                    rlgraph[r][l] += count
        return dictdictize(lrgraph), dictdictize(rlgraph), eojeols

    def extract(self, min_noun_score=0.4, min_frequency=3,
            noun_candidates=None, min_feature_proportion=0.6):

        self._pre_eojeol_analysis(min_frequency)

        if not noun_candidates:
            noun_candidates = [l for l, c in self.lcount.items() if len(l) >= 2]
        noun_candidates = [l for l in noun_candidates
            if self.lcount.get(l, 0) >= min_frequency and not (l in self.r_scores)]

        noun_scores = {}
        for i, l in enumerate(noun_candidates):
            noun_scores[l] = self.predict(l)
            if self.verbose and (i+1) % 1000 == 0:
                message = '\rpredicting noun score ... {} / {}'
                sys.stdout.write(message.format(i+1, len(noun_candidates)))

        if self.verbose:
            print('\rpredicting noun score was done{}'.format(' '*40))

        # debug code
        print('before postprocessing', len(noun_scores))

        noun_scores = self._postprocessing(noun_scores, min_noun_score, min_feature_proportion)

        # debug code
        print('after postprocessing', len(noun_scores))

        self._post_eojeol_analysis(min_frequency)

        for noun in self.noun_dictionary:
            if not (noun in self._noun_scores_postprocessed):
                self.noun_dictionary[noun] = self.predict(noun)

        for noun, score in self._noun_scores_postprocessed.items():
            self.noun_dictionary[noun] = score

        del self._noun_scores_
        del self._noun_scores_postprocessed

        return self.noun_dictionary

    def _pre_eojeol_analysis(self, min_frequency=3, min_eojeol_proportion=0.99):
        def eojeol_to_NV(l, max_eojeol_proportion=0.5, min_noun_score=0.4):
            n = len(l)
            if n < 4: return None
            for e in range(2, n-1):
                (l, r) = (l[:e], l[e:])
                if (self.predict(l)[0] >= min_noun_score) and (r in self._vdictionary):
                    return (l, r)
            return None

        candidates = {l:c for l,c in self.lcount.items()
            if c >= min_frequency and (self.eojeols.get(l, 0) / c) >= min_eojeol_proportion}

        for i, (l, c) in enumerate(candidates.items()):
            if self.verbose and (i+1) % 1000 == 0:
                args = (len(self.noun_dictionary), i+1, len(candidates))
                message = '\rextracting {} nouns using verb/adjective dictionary ... {} / {}'
                sys.stdout.write(message.format(*args))

            nv = eojeol_to_NV(l)
            if not nv:
                continue

            self.noun_dictionary[nv[0]] = self.lcount.get(nv[0], 0)

        if self.verbose:
            message = '\rextracted {} nouns using verb/adjective dictionary'
            sys.stdout.write(message.format(len(self.noun_dictionary)))

    def _post_eojeol_analysis(self, min_frequency=3,
        min_eojeol_proportion=0.99, min_noun_score=0.4):

        candidates = {l:c for l,c in self.lcount.items()
            if c >= min_frequency and (self.eojeols.get(l, 0) / c) >= min_eojeol_proportion}

        begin = len(self.noun_dictionary)
        for i, (l, c) in enumerate(candidates.items()):
            if self.verbose and (i+1) % 1000 == 0:
                args = (len(self.noun_dictionary) - begin, i+1, len(candidates))
                message = '\rextracting {} compounds from eojeols ... {} / {}'
                sys.stdout.write(message.format(*args))

            if l in self._noun_scores_postprocessed:
                continue
            if self._is_NJsubJ(l):
                continue
            if self._is_NJ(l):
                continue
            if self._is_NV(l):
                continue
            if self._hardrule_suffix_filter(l) and self._is_compound(l):
                self.noun_dictionary[l] = c

        if self.verbose:
            message = '\rextracted {} compounds from eojeols'
            sys.stdout.write(message.format(len(self.noun_dictionary) - begin))

    def predict(self, l):
        (norm, score, _total, n_positive_feature, n_feature) = (0, 0, 0, 0, 0)

        for r, frequency in self.lrgraph.get(l, {}).items():
            _total += frequency
            if not r in self.r_scores:
                continue
            norm += frequency
            score += frequency * self.r_scores[r]
            n_feature += 1
            n_positive_feature += 1 if self.r_scores[r] > 0 else 0

        score = score / norm if norm else 0
        n_eojeol = self.eojeols.get(l, 0)
        feature_proportion = norm / (_total - n_eojeol) if (_total - n_eojeol) > 0 else 0
        eojeol_proportion = n_eojeol / _total if _total > 0 else 0
        unique_positive_feature_proportion = 0 if n_feature <= 0 else n_positive_feature / n_feature

        return NewsNounScore(score, _total, feature_proportion, eojeol_proportion, n_positive_feature, unique_positive_feature_proportion)
#         return (score, _total, feature_proportion, eojeol_proportion, n_positive_feature, unique_positive_feature_proportion)

    def _postprocessing(self, noun_scores, min_noun_score=0.4,
        min_feature_proportion=0.6):

        self._noun_scores_ = dict(
            filter(
                lambda x:((x[1][0] > min_noun_score) and
                          (x[1][1] > min_feature_proportion) and
                          (len(x[0]) > 1)),
                       noun_scores.items()
            )
        )

        print('_noun_scores_', len(self._noun_scores_))

        if self.verbose:
            message = 'finding NJsubJ (대학생(으)+로), NsubJ (떡볶+(이)), NVsubE (사기(당)+했다) ... '
            sys.stdout.write(message)

        njsubjs = {l for l in self._noun_scores_ if self._is_NJsubJ(l)}
        nsubs = {l0 for l in self._noun_scores_ for l0 in self._find_NsubJ(l) if not (l in njsubjs)}
        nvsubes = {l for l in self._noun_scores_ if self._is_NVsubE(l) and self._is_NWsub(l) and not self._is_compound(l)}

        if self.verbose:
            sys.stdout.write('done')

        #     unijosa = {}
        self._noun_scores_postprocessed = {}

        for i, (noun, score) in enumerate(self._noun_scores_.items()):
            if self.verbose and (i+1) % 1000 == 0:
                message = '\rchecking hardrules ... {} / {}'
                sys.stdout.write(message.format(i+1, len(self._noun_scores_)))

            if(noun in njsubjs) or (noun in nsubs) or (noun in nvsubes):
                continue
            if not self._hardrule_unijosa_filter(noun) and not self._is_compound(noun):
    #             unijosa[noun] = score
                continue
            if not self._hardrule_suffix_filter(noun):
                continue
            if not self._hardrule_dang_hada_filter(noun):
                continue
            self._noun_scores_postprocessed[noun] = score

        if self.verbose:
            print('\rchecking hardrules ... done')
        return self._noun_scores_postprocessed

    def _is_NJsubJ(self, l, candidate_noun_threshold=0.4, njsub_proportion_threshold=0.8, min_frequency_droprate=0.7):
        """### NJsub + J: 대학생으 + 로"""
        def match_NJsubJ(token):
            for e in l0_candidates:
                if token[e:] in self.josa_dictionary:
                    return True
            return False

        l0_candidates = {l[:e] for e in range(2, len(l))}
        l0_candidates = {len(l0) for l0 in l0_candidates if l0 in self.noun_dictionary or self.predict(l0)[0] > candidate_noun_threshold}
        if not l0_candidates:
            return False
        base = self.l_frequency(l[:max(l0_candidates)])
        if base == 0:
            return False
        elif self.l_frequency(l) / base  > min_frequency_droprate:
            return False
        tokens = {l+r:c for r, c in self.lrgraph.get(l, {}).items()}
        prop = sum((c for token, c in tokens.items() if match_NJsubJ(token))) / sum(tokens.values())
        return prop > njsub_proportion_threshold

    def _find_NsubJ(self, l, candidate_noun_threshold=0.7, nsubj_proportion_threshold=0.7):
        """### Nsub + J: 떡볶 + 이"""
        proportion = lambda l0, l : self.l_frequency(l0) / self.l_frequency(l)
        l0_candidates = {l[:e] for e in range(2, len(l)) if l[e:] in self.josa_dictionary}
        l0_candidates = {l0 for l0 in l0_candidates if (self.predict(l0)[0] > candidate_noun_threshold) and proportion(l0, l) > nsubj_proportion_threshold}
        return l0_candidates

    def _is_NVsubE(self, l, max_eojeol_proportion=0.7):
        """is_NVsubE('성심당') # False
           is_NVsubE('폭행당') # True """

        def eojeol_proportion(w):
            sum_ = sum(self.lrgraph.get(w, {}).values())
            return False if not sum_ else (min(1, self.eojeols.get(w, 0) / sum_) > max_eojeol_proportion)

        r_extensions = self.lrgraph.get(l, {})
        if not r_extensions:
            return False

        n = len(l)
        for b in range(1, 2 if n <= 3 else 3):
            (l0, r0) = (l[:-b], l[-b:])
            if not (l0 in self._noun_scores_) or not (l0 in self.noun_dictionary):
                continue
            r_extension_as_eojeol = sum([eojeol_proportion(r0+r)*c for r, c in r_extensions.items()]) / sum(r_extensions.values())
            if r_extension_as_eojeol > max_eojeol_proportion:
                return True
        return False

    def _is_NWsub(self, l, min_frequency_droprate=0.1):
        def frequency_droprate(l0):
            return 0 if self.l_frequency(l0) <= 0 else (self.l_frequency(l) / self.l_frequency(l0))

        for b in range(1, 2 if len(l) <= 3 else 3):
            (l0, r0) = (l[:-b], l[-b:])
            if (l0 in self._noun_scores_ or l0 in self.noun_dictionary) and (frequency_droprate(l0) < min_frequency_droprate):
                return True
        return False

    def _is_compound(self, l):
        n = len(l)
        if n < 4: return False
        for e in range(2, n):
#             if (e + 1 == n) and (l[:e] in self._noun_scores_):
#                 return True
            (l0, l1) = (l[:e], l[e:])
            if ((l0 in self._noun_scores_) or (l0 in self.noun_dictionary)) and ((l1 in self._noun_scores_) or (l1 in self.noun_dictionary)):
                return True
        if n >= 6:
            for e1 in range(2, n-3):
                for e2 in range(e1 + 2, n-1):
                    (l0, l1, l2) = (l[:e1], l[e1:e2], l[e2:])
                    if ((l0 in self._noun_scores_) or (l0 in self.noun_dictionary)) and ((l1 in self._noun_scores_) or (l1 in self.noun_dictionary)) and ((l2 in self._noun_scores_) or (l2 in self.noun_dictionary)):
                        return True
        return False

    def _is_NJ(self, w):
        for e in range(2, len(w)+1):
            (l, r) = (w[:e], w[e:])
            if ((l in self._noun_scores_) or (l in self.noun_dictionary)) and ((not r) or (self.r_scores.get(r, 0) > 0)):
                return True
        return False

    def _is_NV(self, l):
        for e in range(2, len(l)):
            if (l[:e] in self._noun_scores_postprocessed or l[:e] in self.noun_dictionary) and (l[e:] in self._vdictionary):
                return True
        return False

    def _hardrule_unijosa_filter(self, l, min_frequency=10, max_num_of_josa=1):
        def has_passset(r):
            passset = {'과', '는', '되고', '되는', '되다', '된다', 
                       '들', '들에', '들의', '들이', 
                       '로', '로는', '로도', '로서', '를', 
                       '부터', '뿐', '뿐만', '뿐이', '뿐인', 
                       '에게', '에도', '에서', '에와', '와', 
                       '으로', '으로의', '은', '의', '이', '이나', '이라', '이었', '인', '임', 
                       '처럼', '하다', '한', '할', '했던', '했고', '했다'}        
            return (r in passset)

        if not (l in self._noun_scores_):
            return True
        if self._noun_scores_[l][1] <= min_frequency and self._noun_scores_[l][4] <= max_num_of_josa:
            rdict = self.lrgraph.get(l, {})
            if not rdict:
                return False
            n_passjosa = sum((c for r,c in self.lrgraph[l].items() if (r) and (has_passset(r[:2]) or self._is_NJ(r))))
            n_nonemptyr = sum((c for r, c in self.lrgraph[l].items() if r))
            passset_prop = n_passjosa / n_nonemptyr if n_nonemptyr else 0
            return passset_prop > 0.5
        return True

    def _hardrule_dang_hada_filter(self, l, max_h_proportion=0.5):
        from soynlp.hangle import decompose # TODO check import path
        if not (l[-1] == '당') and (l[:-1] in self._noun_scores_ or l[:-1] in self.noun_dictionary):
            return True
        rdict = self.lrgraph.get(l, {})
        n_base = sum((c for r,c in rdict.items() if c))
        n_h = 0
        for r,c in rdict.items():
            if not r: continue
            rdecompose = decompose(r[0])
            if rdecompose and rdecompose[0] == 'ㅎ':
                n_h += c
        return True if n_base <= 0 else (n_h / n_base < max_h_proportion) 

    def _hardrule_suffix_filter(self, l, min_frequency_droprate=0.8):
        def prop_r(l, r):
            base = sum(self.lrgraph.get(l, {}).values())
            return 0 if base == 0 else self.lrgraph.get(l, {}).get(r, 0) / base

        if l in self._vdictionary:
            return False

        stoplsub = {'갔다', '이는', '겠지', '보는'}
        if l[-2:] in stoplsub: return False
        if (l[-1] == '으') and ((prop_r(l, '로') + prop_r(l, '로써') + prop_r(l, '로만')) + prop_r(l, '로의') > 0.5):
            return False
        if (l[-1] == '지') and (prop_r(l, '만') > 0.5):
            return False
        if (l[-1] == '없') and ((prop_r(l, '는') + prop_r(l, '이') + prop_r(l, '다')) + prop_r(l, '었다') > 0.5):
            return False
        if (l[-1] == '인') or (l[-1] == '은') or (l[-1] == '의') or (l[-1] == '와') or (l[-1] == '과'):
            droprate = 0 if not (l[:-1] in self.lcount) else (self.lcount.get(l, 0) / self.lcount[l[:-1]])
            if droprate < min_frequency_droprate:
                return False
        if (l[-2:] == '들이') or (l[-2:] == '들은') or (l[-2:] == '들도') or (l[-2:] == '들을'):
            droprate = 0 if not (l[:-1] in self.lcount) else (self.lcount.get(l, 0) / self.lcount[l[:-1]])
            if droprate < min_frequency_droprate:
                return False
        if (l[-2:] == '으로') and (len(l) == 3 or (l[:-2] in self.noun_dictionary or l[:-2] in self._noun_scores_postprocessed)):
            return False

        return True

    def l_frequency(self, l):
        return self.lcount.get(l, 0)


from collections import namedtuple
from math import log

NgramScore = namedtuple('NgramScore', 'frequency score')

class Bigram:
    def __init__(self, sentences=None, min_frequency=5, verbose=True, score='frequency',
        filtering_checkpoint=100000, tokenizer=None, ngram_selector=None):

        """
        Attribute:
        ----------
        score : str or functional
            Scoring method. choice in ['frequency', 'pmi', 'mikolov']
        """

        if tokenizer is None:
            tokenizer = lambda x:x.split()

        if ngram_selector is None:
            ngram_selector = lambda x:x

        self.min_frequency = min_frequency
        self.verbose = verbose
        self.score = score
        self.filtering_checkpoint = filtering_checkpoint
        self.tokenizer = tokenizer
        self.ngram_selector = ngram_selector
        self._counter = None

    @property
    def is_trained(self):
        return self._counter

    def train(self, sentences):

        def to_bigram(words):
            bigrams = [(w0, w1) for w0, w1 in zip(words, words[1:])]
            return bigrams

        self._counter = {}

        for i_sent, sent in enumerate(sentences):

            if self.filtering_checkpoint > 0 and i_sent % self.filtering_checkpoint == 0:
                self._counter = {bigram:freq for bigram, freq
                    in self._counter.items() if freq >= self.min_frequency}

            if self.verbose and i_sent % 3000 == 0:
                print('\r[Bigram Extractor] scanning {} bigrams from {} sents'.format(
                    len(self._counter), i_sent), end='', flush=True)

            words = self.tokenizer(sent)
            if len(words) <= 1:
                continue

            bigrams = to_bigram(words)
            for bigram in bigrams:
                self._counter[bigram] = self._counter.get(bigram, 0) + 1

        self._counter = {bigram:freq for bigram, freq
            in self._counter.items() if freq >= self.min_frequency}

        self._unigram = {}
        for bigram, freq in self._counter.items():
            for unigram in bigram:
                self._unigram[unigram] = self._unigram.get(unigram, 0) + 1

        if self.verbose:
            print('\r[Bigram Extractor] scanning {} unigrams, {} bigrams from {} sents'.format(
                len(self._unigram), len(self._counter), i_sent), flush=True)

    def extract(self, topk=-1, threshold=0):
        if self.score == 'frequency':
            return self._extract_by_frequency(topk, threshold)
        elif self.score == 'pmi':
            return self._extract_by_pmi(topk, threshold)
        elif self.score == 'mikolov':
            return self._extract_by_mikolov(topk, threshold)
        raise NotImplemented

    def _extract_by_pmi(self, topk=-1, threshold=0):

        def score(bigram, freq, N):
            base = self._unigram[bigram[0]] * self._unigram[bigram[1]]
            return 0 if base == 0 else log(N * freq / base)

        N = 2 * sum(self._counter.values())
        pmis = {}
        for bigram, freq in self._counter.items():
            pmi = score(bigram, freq, N)
            if pmi >= threshold:
                pmis[bigram] = pmi

        bigrams = {word:NgramScore(freq, pmis[word]) for word, freq
                   in self._counter.items() if word in pmis}
        return bigrams

    def _extract_by_frequency(self, topk=-1, threshold=10):
        bigrams = filter(lambda x:x[1] >= threshold, self._counter.items())
        if topk > 0:
            bigrams = sorted(bigrams, key=lambda x:-x[1])
        bigrams = {word:NgramScore(freq, freq) for word, freq in bigrams}
        return bigrams

    def _extract_by_mikolov(self, topk=-1, threshold=0):

        def score(bigram, freq, N):
            base = self._unigram[bigram[0]] * self._unigram[bigram[1]]
            return 0 if base == 0 else (freq - self.min_frequency) / base

        N = 2 * sum(self._counter.values())
        scores = {}
        for bigram, freq in self._counter.items():
            s = score(bigram, freq, N)
            if s >= threshold:
                scores[bigram] = s

        bigrams = {word:NgramScore(freq, scores[word]) for word, freq
                   in self._counter.items() if word in scores}
        return bigrams


from ._tokenizer import MaxScoreTokenizer

class NounLMatchTokenizer:

    def __init__(self, nouns):
        self._nouns  = nouns

    def __call__(self, sentence, compose_compound=True):
        return self.tokenize(sentence, compose_compound)

    def tokenize(self, sentence, compose_compound=True):

        tokens = [self._max_length_l_tokenize(token)
            for token in sentence.split() if token]

        # remove eojeols which noun does not exist
        tokens = [token for token in tokens if token and token[0]]

        # remove r parts
        tokens = [token[0] for token in tokens]

        if compose_compound:
            tokens = [''.join(token) for token in tokens]
        else:
            tokens = [unit for token in tokens for unit in token]

        tokens = [token for token in tokens if token]
        return tokens

    def _max_length_l_tokenize(self, token):
        
        def nouns_to_larray_and_r(token, nouns_):
            e = sum((len(noun) for noun in nouns_))
            return nouns_, token[e:]

        nouns = []
        n = len(token)

        # string match for generating candidats
        for b in range(n):
            for e in range(b, n+1):
                subword = token[b:e]
                if subword in self._nouns:
                    # (word, begin, length)
                    nouns.append((subword, b, e - b))

        # sort. fisrt order: begin index, second order: length (desc)
        nouns = sorted(nouns, key=lambda x:(x[1], -x[2]))

        nouns_ = []
        e = 0

        while nouns:
            # pop first element
            noun, b, len_ = nouns.pop(0)
            # only concatenate nouns
            if not (b == e):
                return nouns_to_larray_and_r(token, nouns_)
            # append noun and update end index
            nouns_.append(noun)
            e = b + len_
            nouns = [noun for noun in nouns if noun[1] >= e]

        return nouns_to_larray_and_r(token, nouns_)

class NounMatchTokenizer:

    def __init__(self, noun_scores):
        self._tokenizer = MaxScoreTokenizer(scores=noun_scores)

    def __call__(self, sentence, flatten=True, compose_compound=True):
        return self.tokenize(sentence, flatten, compose_compound)

    def tokenize(self, sentence, flatten=True, compose_compound=True):

        def concatenate(eojeol, words):
            words_, b, e, score = [], 0, 0, 0
            for noun_, b_, e_, score_, _ in words:
                if e == b_:
                    e, score = e_, max(score, score_)
                else:
                    words_.append((eojeol[b:e], b, e, score, e-b))
                    b, e = b_, e_
            if e > b:
                words_.append((eojeol[b:e], b, e, score, e-b))
            return words_

        sentence_ = []
        for eojeol in sentence.split():

            eojeol = eojeol.strip()
            if not eojeol:
                continue

            words = self._tokenizer(eojeol, flatten=False)[0]
            # remove non-noun words
            words = [word for word in words if word[3] > 0]

            if compose_compound:
                words = concatenate(eojeol, words)

            sentence_.append(words)

        if flatten:
            sentence_ = [word[0] for words in sentence_ for word in words if word[0]]

        return sentence_


from collections import namedtuple
from soynlp.hangle import decompose
from soynlp.lemmatizer import _lemma_candidate
from soynlp.lemmatizer import _conjugate_stem

EomiScore = namedtuple('EomiScore', 'frequency score')

class EomiExtractor:

    def __init__(self, lrgraph, stems, nouns,
        min_num_of_features=5, verbose=True, logpath=None):

        self.lrgraph = lrgraph
        self._stems = stems
        self._nouns = nouns
        self.min_num_of_features = min_num_of_features
        self.verbose = verbose
        self.logpath = logpath
        self._eomis = None

    @property
    def is_trained(self):
        return self._eomis is not None

    def _print(self, message, replace=False, newline=True):
        header = '[Eomi Extractor]'
        if replace:
            print('\r{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)
        else:
            print('{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)

    def extract(self, condition=None, min_eomi_score=0.3,
        min_frequency=1, reset_lrgraph=True):

        # reset covered eojeol count and extracted eomis
        self._num_of_covered_eojeols = 0
        self._eomis = {}

        self._stem_surfaces = {l for stem in self._stems for l in _conjugate_stem(stem)}

        # base prediction
        candidates = self._candidates_from_stem_surfaces(condition)

        prediction_scores = self._batch_prediction(
            candidates, min_eomi_score, self.min_num_of_features)

        eomi_surfaces = {eomi:score for eomi, score in prediction_scores.items()
            if (score[1] >= min_eomi_score)}

        if self.verbose:
            message = 'eomi lemmatization with {} candidates'.format(len(eomi_surfaces))
            self._print(message, replace=False, newline=True)

        self.lrgraph.reset_lrgraph()
        lemmas = self._eomi_lemmatize(eomi_surfaces)

        lemmas = {eomi:score for eomi, score in lemmas.items()
            if (score[0] >= min_frequency) and (score[1] >= min_eomi_score)}

        lemmas = self._post_processing(lemmas)

        if self.logpath:
            with open(self.logpath+'_eomi_prediction_score.log', 'w', encoding='utf-8') as f:
                f.write('eomi frequency score\n')

                for word, score in sorted(prediction_scores.items(), key=lambda x:-x[1][1]):
                    f.write('{} {} {}\n'.format(word, score[0], score[1]))

        if self.verbose:
            message = '{} eomis extracted with min frequency = {}, min score = {}'.format(
                len(lemmas), min_frequency, min_eomi_score)
            self._print(message, replace=False, newline=True)

        self._check_covered_eojeols(lemmas) # TODO with lemma

        self._eomis = lemmas

        if reset_lrgraph:
            self.lrgraph.reset_lrgraph()

        del self._stem_surfaces

        lemmas_ = {eomi:EomiScore(score[0], score[1]) for eomi, score in lemmas.items()}
        return lemmas_

    def predict(self, r, min_eomi_score=0.3,
        min_num_of_features=5, debug=False):

        features = self.lrgraph.get_l(r, -1)

        pos, neg, unk = self._predict(features, r)

        base = pos + neg
        score = 0 if base == 0 else (pos - neg) / base
        support = pos + unk if score >= min_eomi_score else neg + unk

        features_ = self._refine_features(features, r)
        n_features_ = len(features_)

        if debug:
            print('pos={}, neg={}, unk={}, n_features_={}'.format(
                pos, neg, unk, n_features_))

        if n_features_ >= min_num_of_features:
            return support, score
        else:
            # TODO
            return (0, 0)

    def _predict(self, features, r):

        pos, neg, unk = 0, 0, 0

        for l, freq in features:
            if self._exist_longer_pos(l, r): # ignore
                continue
            if l in self._stem_surfaces:
                pos += freq
            elif self._is_aNoun_Verb(l):
                pos += freq
            elif self._has_stem_at_last(l):
                unk += freq
            else:
                neg += freq

        return pos, neg, unk

    def _exist_longer_pos(self, l, r):
        for i in range(1, len(r)+1):
            if (l + r[:i]) in self._stem_surfaces:
                return True
        return False

    def _is_aNoun_Verb(self, l):
        return (l[0] in self._nouns) and (l[1:] in self._stem_surfaces)

    def _has_stem_at_last(self, l):
        for i in range(1, len(l)):
            if l[-i:] in self._stem_surfaces:
                return True
        return False

    def _refine_features(self, features, r):
        return [(l, count) for l, count in features if
            ((l in self._stem_surfaces) and (not self._exist_longer_pos(l, r)))]

    def _candidates_from_stem_surfaces(self, condition=None):

        def satisfy(word, e):
            return word[-e:] == condition

        # noun candidates from positive featuers such as Josa
        R_from_L = {}

        for l in self._stem_surfaces:
            for r, c in self.lrgraph.get_r(l, -1):

                # candidates filtering for debugging
                # condition is last chars in R
                if not condition:
                    R_from_L[r] = R_from_L.get(r,0) + c
                    continue

                # for debugging
                if satisfy(r, len(condition)):
                    R_from_L[r] = R_from_L.get(r,0) + c

        return R_from_L

    def _batch_prediction(self, eomi_candidates,
        min_eomi_score=0.3, min_num_of_features=5):

        prediction_scores = {}

        n = len(eomi_candidates)
        for i, r in enumerate(sorted(eomi_candidates, key=lambda x:-len(x))):

            if self.verbose and i % 10000 == 9999:
                percentage = '%.2f' % (100 * (i+1) / n)
                message = '  -- batch prediction {} % of {} words'.format(percentage, n)
                self._print(message, replace=True, newline=False)

            # base prediction
            support, score = self.predict(
                r, min_eomi_score, min_num_of_features)
            prediction_scores[r] = (support, score)

            # if their score is higher than min_eomi_score,
            # remove eojeol pattern from lrgraph
            if score >= min_eomi_score:
                for l, count in self.lrgraph.get_l(r, -1):
                    if ((l in self._stem_surfaces) or
                        self._is_aNoun_Verb(l)):
                        self.lrgraph.remove_eojeol(l+r, count)

        self.lrgraph.reset_lrgraph()

        if self.verbose:
            message = 'batch prediction was completed for {} words'.format(n)
            self._print(message, replace=True, newline=True)

        return prediction_scores

    def _eomi_lemmatize(self, eomis):

        def merge_score(freq0, score0, freq1, score1):
            return (freq0 + freq1, (score0 * freq0 + score1 * freq1) / (freq0 + freq1))

        lemmas = {}
        for eomi, (_, score0) in eomis.items():
            for stem_surface, count in self.lrgraph.get_l(eomi, -1):
                try:
                    for stem, lemma in _lemma_candidate(stem_surface, eomi):
                        if not (stem in self._stems):
                            continue
                        lemmas[lemma] = merge_score(count, score0, *lemmas.get(lemma, (0, 0)))
                # stem 이 한글이 아닌 경우 불가
                except Exception as e:
                    continue

        return lemmas

    def _post_processing(self, eomis):
        eomis_ = {}
        for eomi, score in eomis.items():
            # 어미의 첫 종성이 ㅎ 으로 끝나는 경우
            if decompose(eomi[0])[2] == 'ㅎ':
                continue
            # TODO
            # Remove E + V + E : -서가지고
            # Remove V + E : -싶구나
            eomis_[eomi] = score

        return eomis_

    def _check_covered_eojeols(self, eomis):
        # TODO
        return None


# -*- encoding:utf8 -*-

from collections import defaultdict
from collections import namedtuple
import math
import numpy as np
import pickle
import sys
from soynlp.utils import get_process_memory

Scores = namedtuple('Scores', 'cohesion_forward cohesion_backward left_branching_entropy right_branching_entropy left_accessor_variety right_accessor_variety leftside_frequency rightside_frequency')

def _entropy(dic):
    if not dic: 
        return 0.0
    sum_ = sum(dic.values())
    entropy = 0
    for freq in dic.values():
        prob = float(freq) / sum_
        entropy += prob * math.log(prob)
    return -1 * entropy


class WordExtractor:
    
    def __init__(self, sents=None, max_left_length=10, max_right_length=6,
                min_frequency=5, verbose_points=100000,
                min_cohesion_forward=0.05, min_cohesion_backward=0.0, 
                max_droprate_cohesion=0.98, max_droprate_leftside_frequency=0.98,
                min_left_branching_entropy=0.0, min_right_branching_entropy=0.0,
                min_left_accessor_variety=0, min_right_accessor_variety=0,
                remove_subwords=True):
        self.max_left_length = max_left_length
        self.max_right_length = max_right_length
        self.min_frequency = min_frequency
        self.L = {}
        self.R = {}
        self._aL = {}
        self._aR = {}
        self.verbose = verbose_points

        self.min_cohesion_forward = min_cohesion_forward
        self.min_cohesion_backward = min_cohesion_backward
        #self.max_droprate_cohesion = max_droprate_cohesion
        self.max_droprate_leftside_frequency = max_droprate_leftside_frequency
        self.min_left_branching_entropy = min_left_branching_entropy
        self.min_right_branching_entropy = min_right_branching_entropy
        self.min_left_accessor_variety = min_left_accessor_variety
        self.min_right_accessor_variety = min_right_accessor_variety
        self.remove_subwords = remove_subwords
        
        if sents:
            self.train(sents)
        
    def train(self, sents, num_for_pruning = 0, cumulate=True):
        def prune_extreme_case():
            self.L = defaultdict(lambda: 0, {w:f for w,f in self.L.items() if f >= self.min_frequency})
            self.R = defaultdict(lambda: 0, {w:f for w,f in self.R.items() if f >= self.min_frequency})
        def prune_extreme_case_a():
            self._aL = defaultdict(lambda: 0, {w:f for w,f in self._aL.items() if f > 1})
            self._aR = defaultdict(lambda: 0, {w:f for w,f in self._aR.items() if f > 1})

        if cumulate:
            self.L = defaultdict(int, self.L)
            self.R = defaultdict(int, self.R)
            self._aL = defaultdict(int, self._aL)
            self._aR = defaultdict(int, self._aR)
        else:
            self.L = defaultdict(int)
            self.R = defaultdict(int)
            self._aL = defaultdict(int)
            self._aR = defaultdict(int)

        for num_sent, sent in enumerate(sents):
            if sys.version_info.major == 2:
                words = map(unicode, sent.strip().split())
            else:
                words = sent.split()

            for word in words:
                if (not word) or (len(word) <= 1):
                    continue
                word_len = len(word)
                for i in range(1, min(self.max_left_length + 1, word_len)+1):
                    self.L[word[:i]] += 1
                for i in range(1, min(self.max_right_length + 1, word_len)):
                    self.R[word[-i:]] += 1

            if len(words) <= 1:
                continue
            for left_word, word, right_word in zip([words[-1]]+words[:-1], words, words[1:]+[words[0]]):
                self._aL['%s %s' % (word, right_word[0])] += 1
                self._aR['%s %s' % (left_word[-1], word)] += 1

                word_len = len(word)
                for i in range(1, min(self.max_right_length + 1, word_len)):
                    self._aL['%s %s' % (word[-i:], right_word[0])] += 1
                for i in range(1, min(self.max_left_length + 1, word_len)):
                    self._aR['%s %s' % (left_word[-1], word[:i])] += 1

            if (num_for_pruning > 0) and ( num_sent % num_for_pruning == 0):
                prune_extreme_case()
            if (self.verbose > 0) and ( num_sent % self.verbose == 0):
                sys.stdout.write('\rtraining ... (%d in %d sents) use memory %.3f Gb' % (num_sent, len(sents), get_process_memory()))

        prune_extreme_case()
        prune_extreme_case_a()
        if (self.verbose > 0):
            print('\rtraining was done. used memory %.3f Gb' % (get_process_memory()))
        self.L = dict(self.L)
        self.R = dict(self.R)
        self._aL = dict(self._aL)
        self._aR = dict(self._aR)

    def extract(self, scores=None):
        if not scores:
            scores = self.word_scores()
        scores_ = {}
        for word, score in sorted(scores.items(), key=lambda x:len(x[0])):
            if len(word) <= 2:
                scores_[word] = score
                continue
            if (score.cohesion_forward < self.min_cohesion_forward) or \
                (score.cohesion_backward < self.min_cohesion_backward) or \
                (score.left_branching_entropy < self.min_left_branching_entropy) or \
                (score.right_branching_entropy < self.min_right_branching_entropy) or \
                (score.left_accessor_variety < self.min_left_accessor_variety) or \
                (score.right_accessor_variety < self.min_right_accessor_variety) or \
                (max(score.leftside_frequency, score.rightside_frequency) < self.min_frequency):
                continue
            scores_[word] = score
            if not self.remove_subwords:
                continue
            subword = word[:-1]            
            droprate_leftside_frequency = 0 if not (subword in self.L) else score.leftside_frequency / self.L[subword]
            if (droprate_leftside_frequency > self.max_droprate_leftside_frequency) and (subword in scores_):
                del scores_[subword]
        return scores_
    
    def word_scores(self):
        cps = self.all_cohesion_scores()
        bes = self.all_branching_entropy()
        avs = self.all_accessor_variety()
        scores = {}
        for word in self.words():
            cp = cps.get(word, (0, 0))
            be = bes.get(word, (0, 0))
            av = avs.get(word, (0, 0))
            scores[word] = Scores(cp[0], cp[1], be[0], be[1], av[0], av[1], self.L.get(word, 0), self.R.get(word, 0))
        return scores
    
    def all_cohesion_scores(self):
        cps = {}
        words = self.words()
        for i, word in enumerate(words):
            if (self.verbose > 0) and (i % self.verbose == 0):
                sys.stdout.write('\r cohesion probabilities ... (%d in %d)' % (i+1, len(words)))
            cp = self.cohesion_score(word)
            if (cp[0] == 0) and (cp[1] == 0):
                continue
            cps[word] = cp
        if (self.verbose > 0):
            print('\rall cohesion probabilities was computed. # words = %d' % len(cps))
        return cps

    def cohesion_score(self, word):
        word_len = len(word)
        if (not word) or (word_len <= 1):
            return (0, 0)
        l_freq, r_freq = map(float, self.frequency(word))
        l_cohesion = 0 if l_freq == 0 else np.power( (l_freq / self.L[word[0]]), (1 / (word_len - 1)) )
        r_cohesion = 0 if r_freq == 0 else np.power( (r_freq / self.R[word[-1]]), (1 / (word_len - 1)) )
        return (l_cohesion, r_cohesion)
    
    def frequency(self, word):
        return (self.L.get(word, 0), self.R.get(word, 0))
    
    def all_branching_entropy(self, get_score=_entropy):
        def parse_left(extension):
            return extension[:-1]
        def parse_right(extension):
            return extension[1:]
        def sort_by_length(counter):
            sorted_by_length = defaultdict(lambda: [])
            for w in counter.keys():
                sorted_by_length[len(w)].append(w)
            return sorted_by_length
        def get_entropy_table(parse, sorted_by_length, sorted_by_length_a, max_length, counter, counter_a):
            num_sum = sum((len(words) for length, words in sorted_by_length.items()))
            be = {}
            for word_len in range(2, max_length):
                words = sorted_by_length.get(word_len, [])
                extensions = defaultdict(lambda: [])
                for word in words:
                    extensions[parse(word)].append(word)
                words_ = sorted_by_length_a.get(word_len+1, [])
                for word in words_:
                    extensions[parse(word.replace(' ',''))].append(word)
                for root_word, extension_words in extensions.items():
                    extension_frequency = {ext:counter_a.get(ext) if ' ' in ext else counter.get(ext) for ext in extension_words}
                    be[root_word] = get_score(extension_frequency)
            return be
        def merge(be_l, be_r):
            be = {word:(v, be_r.get(word, 0)) for word, v in be_l.items()}
            for word, v in be_r.items():
                if word in be_l: continue
                be[word] = (0, v)
            return be

        be_l = get_entropy_table(parse_right, sort_by_length(self.R), sort_by_length(self._aR), self.max_right_length+1, self.R, self._aR)
        be_r = get_entropy_table(parse_left, sort_by_length(self.L), sort_by_length(self._aL), self.max_left_length+1, self.L, self._aL)
        be = merge(be_l, be_r)
        if self.verbose > 0:
            print_head = 'branching entropies' if get_score == _entropy else 'accessor variety'
            print('\rall %s was computed # words = %d' % (print_head, len(be)))
        return be

    def branching_entropy(self, word):
        word_len = len(word)
        lsb = { w:f for w,f in self.R.items() if ((len(w) - 1) == word_len) and (w[1:] == word) }
        lsb.update({ w:f for w,f in self._aR.items() if ((len(w) - 2) == word_len) and (w[2:] == word) })
        rsb = { w:f for w,f in self.L.items() if ((len(w) - 1) == word_len) and (w[:-1] == word) }
        rsb.update({ w:f for w,f in self._aL.items() if ((len(w) - 2) == word_len) and (w[:-2] == word) })
        be_l = 0 if not lsb else _entropy(lsb)
        be_r = 0 if not rsb else _entropy(rsb)
        return (be_l, be_r)

    def all_accessor_variety(self):
        return self.all_branching_entropy(get_score=len)

    def accessor_variety(self, word):
        word_len = len(word)
        lsb = { w:f for w,f in self.R.items() if ((len(w) - 1) == word_len) and (w[1:] == word) }
        lsb.update({ w:f for w,f in self._aR.items() if ((len(w) - 2) == word_len) and (w[2:] == word) })
        rsb = { w:f for w,f in self.L.items() if ((len(w) - 1) == word_len) and (w[:-1] == word) }
        rsb.update({ w:f for w,f in self._aL.items() if ((len(w) - 2) == word_len) and (w[:-2] == word) })
        av_l = 0 if lsb == False else len(lsb)
        av_r = 0 if rsb == False else len(rsb)
        return (av_l, av_r)

    def words(self):
        words = {word for word in self.L.keys() if len(word) <= self.max_left_length}
        words.update({word for word in self.R.keys() if len(word) <= self.max_right_length})
        return words

    def save(self, fname):
        configuration = {
            'max_left_length': self.max_left_length,
            'max_right_length': self.max_right_length,
            'min_frequency': self.min_frequency,
            'verbose_points': self.verbose,
            'min_cohesion_forward': self.min_cohesion_forward,
            'min_cohesion_backward': self.min_cohesion_backward,
            #'max_droprate_cohesion': self.max_droprate_cohesion,
            'max_droprate_leftside_frequency': self.max_droprate_leftside_frequency,
            'min_left_branching_entropy': self.min_left_branching_entropy,
            'min_right_branching_entropy': self.min_right_branching_entropy,
            'min_left_accessor_variety': self.min_left_accessor_variety,
            'min_right_accessor_variety': self.min_right_accessor_variety,
            'remove_subwords': self.remove_subwords
        }
        data = {
            'L': self.L,
            'R': self.R,
            'aL': self._aL,
            'aR': self._aR
        }
        params = {
            'configuration': configuration,
            'data': data
            }
        with open(fname, 'wb') as f:
            pickle.dump(params, f)

    def load(self, fname):
        with open(fname, 'rb') as f:
            params = pickle.load(f)

        configuration = params['configuration']
        self.max_left_length = configuration['max_left_length']
        self.max_right_length = configuration['max_right_length']
        self.min_frequency = configuration['min_frequency']
        self.verbose = configuration['verbose_points']

        self.min_cohesion_forward = configuration['min_cohesion_forward']
        self.min_cohesion_backward = configuration['min_cohesion_backward']
        #self.max_droprate_cohesion = configuration['max_droprate_cohesion']
        self.max_droprate_leftside_frequency = configuration['max_droprate_leftside_frequency']
        self.min_left_branching_entropy = configuration['min_left_branching_entropy']
        self.min_right_branching_entropy = configuration['min_right_branching_entropy']
        self.min_left_accessor_variety = configuration['min_left_accessor_variety']
        self.min_right_accessor_variety = configuration['min_right_accessor_variety']
        self.remove_subwords = configuration['remove_subwords']

        data = params['data']
        self.L = data['L']
        self.R = data['R']
        self._aL = data['aL']
        self._aR = data['aR']

        del params
        del configuration
        del data



from .utils import get_available_memory
from .utils import get_process_memory
from .utils import sort_by_alphabet
from .utils import check_dirs
from .utils import DoublespaceLineCorpus
from .utils import EojeolCounter
from .utils import LRGraph

__all__ = [
    # utils
    'get_available_memory', 'get_process_memory', 'check_dirs'
    'sort_by_alphabet', 'DoublespaceLineCorpus',
    'EojeolCounter', 'LRGraph',
]


# -*- encoding:utf8 -*-

class BaseEvaluator:
    def evaluate(self, candidate):
        # Develop your evaluation functions
        return 0
    def select_best(self, candidates):
        if not candidates:
            return None
        scored = [(c, self.evaluate(c)) for c in candidates]
        best = sorted(scored, key=lambda x:-x[1])[0][0]
        return best

class SimpleEojeolEvaluator(BaseEvaluator):
    def __init__(self, weights=None):
        if not weights:
            weights = (
                ('num_nouns', -0.1),
                ('num_words', -0.15),
                ('has_noun', 0.2),
                ('has_unknown', -0.3),
                ('is_adverb_repeat', 0.1),
                ('is_compound_noun', -0.05),
                ('is_noun_josa', 0.1),
                ('is_a_noun', 0.2)
            )
        self.weights = weights
        
    def evaluate(self, candidate):
        num_nouns = len([True for lr in candidate if lr.l_tag == 'Noun'])
        num_words = len(candidate)
        has_noun = num_nouns > 0
        has_unknown = len([True for lr in candidate if lr.l_tag is None])
        is_adverb_repeat = (num_words > 1 and 
            num_words == len([True for lr in candidate if lr.l_tag == 'Adverb']))
        is_compound_noun = (num_words > 1 and
            num_words == len([True for lr in candidate if lr.l_tag == 'Noun']))
        is_noun_josa = (num_words == 1 and
            (candidate[0].l_tag == 'Noun' and candidate[0].r_tag == 'Josa'))
        is_a_noun = num_words == 1 and candidate[0].l_tag == 'Noun' and not candidate[0].r
        
        score = (num_nouns * self.weights[0][1]
                 + num_words * self.weights[1][1]
                 + has_noun * self.weights[2][1]
                 + has_unknown * self.weights[3][1]
                 + is_adverb_repeat * self.weights[4][1]
                 + is_compound_noun * self.weights[5][1]
                 + is_noun_josa * self.weights[6][1]
                 + is_a_noun * self.weights[7][1])

        return score

class LREvaluator(BaseEvaluator):
    def __init__(self, weights=None, preference=None):
        """
        Arguments
        ---------
        preference: dict[str][str] = float
            Word preference. dict[tag][word] = preference_score
        """
        if not weights:
            weights = (
                ('is_noun_phrase', 0.3),
                ('lr_are_known', 0.15),
                ('is_verb', 0.2),
                ('is_adjective', 0.1),
                ('l_len', 0.1),
                ('r_len', 0.1),
                ('l_is_a_syllable', -0.2),
            )
        self.weights = weights
        if not preference:
            preference = {}
        self.preference = preference

    def select_best(self, candidates):
        scored_candidates = [(c, self.evaluate(c)) for c in candidates]
        best = self._remove_overlapped(scored_candidates)
        return sorted(best, key=lambda x:x[2])
    
    def evaluate(self, candidate):
        is_noun_phrase = candidate.l_tag == 'Noun'
        lr_are_known = candidate.l_tag is not None and candidate.r_tag is not None
        # tag preference
        is_verb = candidate.l_tag == 'Adverb'
        is_adjective = candidate.l_tag == 'Adjective'
        # length
        l_len = candidate.m - candidate.b
        r_len = candidate.e - candidate.m
        l_is_a_syllable = l_len == 1
        # user-defined preference
        preference = 0
        if self.preference:
            preference += self.preference.get(candidate.l_tag, {}).get(candidate.l, 0)
            preference += self.preference.get(candidate.r_tag, {}).get(candidate.r, 0)

        score = (is_noun_phrase * self.weights[0][1]
                 + lr_are_known * self.weights[1][1]
                 + is_verb * self.weights[2][1]
                 + is_adjective * self.weights[3][1]
                 + l_len * self.weights[4][1]
                 + r_len * self.weights[5][1]
                 + l_is_a_syllable * self.weights[6][1]
                 + preference
                )

        return score

    def _remove_overlapped(self, scored_candidates):
        best = []
        sorted_ = sorted(scored_candidates, key=lambda x:-x[1])
        
        while sorted_:
            best.append(sorted_.pop(0)[0])
            (b, e) = (best[-1].b, best[-1].e)
            
            # Find overlapped 
            removals = [i for i, (c, _) in enumerate(sorted_) if b < c.e and e > c.b]
            
            for idx in reversed(removals):
                del sorted_[idx]
                
        return sorted(best, key=lambda x:x[2])


import os
from collections import Counter
from scipy.sparse import csr_matrix

class BaseVectorizer:

    def __init__(self, tokenizer=lambda x:x.split(), min_tf=0,
        max_tf=99999999, min_df=0, max_df=1.0, stopwords=None,
        lowercase=True, verbose=True):

        assert 0 <= min_df < 1
        assert 0 < max_df <= 1

        self.tokenizer = tokenizer
        self.min_tf = min_tf
        self.max_tf = max_tf
        self.min_df = min_df
        self.max_df = max_df
        self.stopwords = stopwords if stopwords else {}
        self.lowercase = lowercase
        self.verbose = verbose

        self._check_points = 500

        self.vocabulary_ = {}
        self.idx2vocab = []
        self.n_vocabs = 0

    def fit_transform(self, docs):
        self.fit(docs)
        return self.transform(docs)

    def fit(self, docs):
        # counting
        df = {}
        tf = {}

        for i_doc, doc in enumerate(docs):
            if self.verbose and i_doc % self._check_points == 0:
                print('\rscanned {} docs'.format(i_doc), flush=True, end='')

            counter = Counter((token for token in self.tokenizer(doc)))
            for term, freq in counter.items():
                df[term] = df.get(term, 0) + 1
                tf[term] = tf.get(term, 0) + 1

        if self.verbose:
            print('\rscanning was done{}'.format(' '*40), flush=True)

        # filtering
        n_docs = i_doc + 1
        min_df = int(n_docs * self.min_df)
        max_df = int(n_docs * self.max_df)
        df = {term:df_t for term, df_t in df.items() if min_df <= df_t <= max_df}
        tf = {term:tf_t for term, tf_t in tf.items() if self.min_tf <= tf_t <= self.max_tf}

        # build vocabulary_
        vocabs = {term:tf_t for term, tf_t in tf.items() if term in df}
        self.vocabulary_ = {term:idx for idx, (term, _) in enumerate(
            sorted(vocabs.items(), key=lambda x:-x[1]))}
        self.idx2vocab = [term for term, _ in sorted(
            self.vocabulary_.items(), key=lambda x:x[1])]
        self.n_vocabs = len(self.idx2vocab)
        
        if self.verbose:
            print('{} terms are recognized'.format(self.n_vocabs), flush=True)

        return self

    def transform(self, docs):
        rows = []
        cols = []
        data = []
        for i_doc, doc in enumerate(docs):
            if self.verbose and i_doc % self._check_points == 0:
                print('\rtransformed {} docs'.format(i_doc), flush=True, end='')

            bow = self.encode_a_doc_to_bow(doc)
            for term, count in bow.items():
                rows.append(i_doc)
                cols.append(term)
                data.append(count)

        if self.verbose:
            print('\rtransforming docs to term frequency marix was done', flush=True)

        return csr_matrix((data, (rows, cols)), shape=(i_doc+1, self.n_vocabs))

    def fit_to_file(self, docs, file_path, encoding='utf-8'):
        self.fit(docs)
        self.to_file(docs, file_path, encoding)

    def to_file(self, docs, file_path, encoding='utf-8'):
        # check n_elements
        n_elements = 0
        for i, doc in enumerate(docs):
            if self.verbose and i % self._check_points == 0:
                print('\rscanning number of elements from {} docs'.format(
                    i), flush=True, end='')
            words = self.tokenizer(doc)
            n_elements += len({word for word in words if word in self.vocabulary_})
        n_docs = i + 1
        if self.verbose:
            print('\rscanning number of elements was done. from {} docs'.format(
                n_docs), flush=True, end='')

        # directory check
        directory = os.path.dirname(file_path)
        if not os.path.exists(directory):
            os.makedirs(directory)

        # to file
        with open(file_path, 'w', encoding=encoding) as f:
            # header
            f.write('%%MatrixMarket matrix coordinate integer general\n')
            f.write('%\n')
            f.write('{} {} {}\n'.format(n_docs, self.n_vocabs, n_elements))
            # doc term frequency
            for i, doc in enumerate(docs):
                if self.verbose and i % self._check_points == 0:
                    print('\rwriting to file {} % {}'.format(
                        100 * i / n_docs, ' '*30), flush=True, end='')
                words = self.tokenizer(doc)
                words = Counter([self.vocabulary_[word]
                    for word in words if word in self.vocabulary_])
                for j, count in words.items():
                    f.write('{} {} {}\n'.format(i+1, j+1, count))
        if self.verbose:
            print('\rwriting to file was done. {} docs'.format(n_docs), flush=True)

    def __len__(self):
        return self.n_vocabs

    def encode_a_doc_to_list(self, doc):
        return [self.vocabulary_[term] for term in self.tokenizer(doc) if term in self.vocabulary_]

    def decode_from_list(self, doc):
        return [self.idx2vocab[idx] for idx in doc if 0 <= idx < self.n_vocabs]

    def encode_a_doc_to_bow(self, doc):
        bow = Counter(self.tokenizer(doc))
        bow = {self.vocabulary_[term]:count for term, count in bow.items() if term in self.vocabulary_}
        return bow

    def decode_from_bow(self, bow):
        bow = {self.idx2vocab[idx]:count for idx, count in bow.items() if 0 <= idx < self.n_vocabs}
        return bow

    def save(self, fname):
        if fname[-6:] != '.vocab':
            fname += '.vocab'
        with open(fname, 'w', encoding='utf-8') as f:
            for vocab in self.idx2vocab:
                f.write('{}\n'.format(vocab))
    
    def load(self, fname):
        if fname[-6:] != '.vocab':
            fname += '.vocab'
        with open(fname, encoding='utf-8') as f:
            self.idx2vocab = [term.strip() for term in f]
        self.vocabulary_ = {term:idx for idx, term in enumerate(self.idx2vocab)}
        self.n_vocabs = len(self.idx2vocab)

    def vocabs(self):
        return [term for term in sorted(self.vocabulary_, key=lambda x:self.vocabulary_[x])]
    
    def _set_vocabulary(self, vocabulary_list):
        self.idx2vocab = vocabulary_list
        self.vocabulary_ = {v:i for i,v in enumerate(self.idx2vocab)}
        self.n_vocabs = len(self.idx2vocab)


# -*- encoding:utf8 -*-
import sys
from collections import namedtuple

# b:begin, m: middle, e: end
LR = namedtuple('LR', 'l l_tag r r_tag b m e')

class BaseTemplateMatcher:
    def generate(self, token):
        raise NotImplementedError

class EojeolTemplateMatcher(BaseTemplateMatcher):
    def __init__(self, dictionary, single_tags=None, lr_templates=None):
        if not single_tags:
            single_tags = ['Noun', 'Verb', 'Adjective', 'Adverb', 'Exclamation']
        if not lr_templates:
            lr_templates = [('Noun', 'Verb'), ('Noun', 'Adjective'), ('Noun', 'Josa')]
        self.dictionary = dictionary
        self.single_tags = single_tags
        self.lr_templates = lr_templates

    def generate(self, eojeol):
        n = len(eojeol)
        candidates = []
        for tag in self.dictionary.get_pos(eojeol):
            if tag in self.single_tags:
                candidates.append([LR(eojeol, tag, '', None, 0, n, n)])
        if not candidates:
            candidates.append([LR(eojeol, None, '', None, 0, n, n)])

        for b in range(1, n):
            l, r = eojeol[:b], eojeol[b:]
            for l_tag, r_tag in self.lr_templates:
                if self.dictionary.word_is_tag(l, l_tag) and self.dictionary.word_is_tag(r, r_tag):
                    candidates.append([LR(l, l_tag, r, r_tag, 0, b, n)])
        
        compound_noun = self._decompose_compound(eojeol, 'Noun')
        if compound_noun:
            candidates.append(compound_noun)
            
        compound_adverb = self._decompose_compound(eojeol, 'Adverb')
        if compound_adverb:
            candidates.append(compound_adverb)

        return candidates

    def _decompose_compound(self, eojeol, tag):
        n, b = len(eojeol), 0
        words = []
        while b < n:
            e = b + 1
            next_round = False
            for i in range(e, n+1):
                if next_round or (b == 0 and i == n):
                    break
                subword = eojeol[b:i]
                if self.dictionary.word_is_tag(subword, tag):
                    words.append(LR(subword, tag, '', None, b, i, i))
                    b = i
                    next_round = True
                    break
            if not next_round:
                return []
        return words

class LRTemplateMatcher(BaseTemplateMatcher):
    def __init__(self, dictionary, ltags=None, templates=None):
        if not ltags:
            ltags = {'Noun', 'Adjective', 'Verb', 'Adverb', 'Exclamation'}
        if not templates:
            templates = {'Noun': ('Josa', 'Verb', 'Adjective')}
        
        self.dictionary = dictionary
        self.ltags = ltags
        self.rtags = {tag for tags in templates.values() for tag in tags}
        self.templates = templates
            
    def generate(self, token):
        if sys.version_info.major == 2:
            token = unicode(token)
        candidates = self._initialize_L(token)
        candidates = self._expand_R(token, candidates)
        return candidates
    
    def _pos_L(self, word):
        poses = self.dictionary.get_pos(word)
        poses = {pos for pos in poses if pos in self.ltags}
        return poses

    def _initialize_L(self, t):
        n = len(t)
        candidates = []
        
        for b in range(n):
            for e in range(b+2, min(n, b + self.dictionary.max_length) + 1):
                l = t[b:e]
                l_tags = self._pos_L(l)
                
                if not l_tags:
                    continue
                    
                for l_tag in l_tags:
                    candidates.append([l, l_tag, b, e])
                    
        candidates = self._remove_subset_l(candidates)
        return sorted(candidates, key=lambda x:x[2])

    def _remove_subset_l(self, candidates):
        candidates_ = []
        for l_tag in self.ltags:
            
            # Sort by length of L
            sorted_ = sorted(
                filter(lambda x:x[1] == l_tag, candidates), 
                key=lambda x:-(x[3] - x[2])
            )
            
            while sorted_:
                candidates_.append(sorted_.pop(0))
                (b, e) = (candidates_[-1][2], candidates_[-1][3])
                
                 # Find overlapped
                removals = [i for i, c in enumerate(sorted_) if b <= c[2] and e >= c[3]]
                
                for idx in reversed(removals):
                    del sorted_[idx]
                    
        return candidates_

    def _expand_R(self, t, candidates):
        n = len(t)
        expanded = []

        for (l, l_tag, b, e1) in candidates:
            last = min(self.dictionary.max_length + e1, n)
            
            for e2 in range(e1, last + 1):
                r = t[e1:e2]
                
                if not r:
                    expanded.append(LR(l, l_tag, r, None, b, e1, e2))
                else:
                    for r_tag in self.templates.get(l_tag, []):
                        if not self.dictionary.word_is_tag(r, r_tag):
                            continue
                        expanded.append(LR(l, l_tag, r, r_tag, b, e1, e2))
                        
        expanded = self._remove_subset_r(expanded)
        return sorted(expanded, key=lambda x:x.b)

    def _remove_subset_r(self, expanded):
        expanded_ = []
        rtags = [rtag for rtag in self.rtags] + [None]
        for r_tag in rtags:
            
            # Sory by length of R
            sorted_ = sorted(
                filter(lambda x:x.r_tag == r_tag, expanded),
                key=lambda x:-(x.e - x.m)
            )
            
            while sorted_:
                expanded_.append(sorted_.pop(0))
                
                l_tag = expanded_[-1].l_tag
                (b, m, e) = expanded_[-1][-3:]
                
                 # Find subset
                removals = [i for i, c in enumerate(sorted_) 
                            if ((c.l_tag == l_tag) 
                                and (b == c.b and m == c.m)
                                and (m <= c.m and e >= c.e))
                           ]

                for idx in reversed(removals):
                    del sorted_[idx]

        return expanded_


from ._tokenizer import LTokenizer
from ._tokenizer import MaxScoreTokenizer
from ._tokenizer import MaxLRScoreTokenizer
from ._tokenizer import RegexTokenizer
from ._normalizer import normalize
from ._noun_tokenizer import NounLMatchTokenizer
from ._noun_tokenizer import NounMatchTokenizer


import math

def extract_domain_pos_features(prediction_scores, lrgraph,
    known_pos_features, ignore_features=None,
    min_noun_score=0.3, min_noun_frequency=100,
    min_pos_score=0.3, min_pos_feature_frequency=1000,
    min_num_of_unique_lastchar=4, min_entropy_of_lastchar=0.5,
    min_noun_entropy=1.5):

    nouns = {noun for noun, score in prediction_scores.items()
             if ((score[0] >= min_noun_score) 
                 and (score[1] >= min_noun_frequency))}

    if ignore_features is None:
        ignore_features = {}

    pos_candidates = {}
    for noun in nouns:
        for r, count in lrgraph.get_r(noun, -1):
            if (r in known_pos_features) or (r in ignore_features):
                continue
            pos_candidates[r] = pos_candidates.get(r, 0) + count

    # 1st. frequency filtering (ignoring L is noun)
    pos_candidates = {r:count for r, count in pos_candidates.items() 
                      if count >= min_pos_feature_frequency}

    # add known pos features for unknown feature prediction
    domain_pos_features = {r:None for r in known_pos_features}

    # from shorter to longer
    for r in sorted(pos_candidates, key=lambda x:len(x)):

        if (r in known_pos_features) or (r in nouns):
            continue

        features = _get_noun_feature(r, lrgraph)
        score, freq = predict(r, features, nouns, domain_pos_features,
            min_pos_score, min_pos_feature_frequency,
            min_num_of_unique_lastchar, min_entropy_of_lastchar)

        # noun entropy
        noun_sum = sum((c for l, c in features if l in nouns))
        noun_entropy = [c/noun_sum for l, c in features if l in nouns]
        noun_entropy = sum([-math.log(p) * p for p in noun_entropy])

        if ((score >= min_pos_score) and
            (freq >= min_pos_feature_frequency) and
            (noun_entropy >= min_noun_entropy)):
            domain_pos_features[r] = (score, freq)

    # remove known features
    domain_pos_features = {r:score for r, score in domain_pos_features.items()
                           if not (r in known_pos_features)}

    return domain_pos_features

def _get_noun_feature(r, lrgraph):
    return [(l,c) for l, c in lrgraph.get_l(r, -1) if len(l) > 1]

def predict(r, features, nouns, pos_r,
    min_pos_score=0.3, min_pos_feature_frequency=30,
    min_num_of_unique_lastchar=4, min_entropy_of_lastchar=0.5):

    n_unique, n_sum, entropy = _last_character_criterions(features)
    if ((n_unique < min_num_of_unique_lastchar) or 
        (entropy < min_entropy_of_lastchar)):
        return (0, 0)

    pos, neg, unk = _predict(r, features, nouns, pos_r)
    score = (pos - neg) / (pos + neg) if (pos + neg) > 0 else 0
    freq = pos if score >= min_pos_score else neg + unk

    if freq < min_pos_feature_frequency :
        return (0, freq)
    else:
        return (score, freq)

def _predict(r, features, nouns, pos_r):

    pos, neg, unk = 0, 0, 0

    for l, freq in features:
        if len(l) <= 1:
            continue
        if _is_NJ(r, nouns, pos_r):
            neg += freq
        elif _exist_longer_noun(l, r, nouns):
            neg += freq
        elif l in nouns:
            pos += freq
        else:
            unk += freq

    return pos, neg, unk

def _exist_longer_noun(l, r, nouns):

    for i in range(1, len(r)+1):
        if (l + r[:i]) in nouns:
            return True
    return False

def _is_NJ(r, nouns, pos_r):

    n = len(r)
    for i in range(1, n):
        if r[:i] in nouns:
            for j in range(i, n):
                return r[j:] in pos_r
    return False

def _last_character_criterions(lcount):

    counter = {}
    for l, count in lcount:
        last = l[-1]
        counter[last] = counter.get(last, 0) + count

    n_unique = len(counter)
    n_sum = sum(counter.values())
    entropy = [freq/n_sum for freq in counter.values()]
    entropy = -1 * sum((p * math.log(p) for p in entropy))

    return n_unique, n_sum, entropy


# -*- encoding:utf8 -*-

from collections import OrderedDict
from collections import namedtuple
from math import log

from soynlp.tokenizer import MaxScoreTokenizer
from ._dictionary import Dictionary

default_profile= OrderedDict([
        ('cohesion_l', 0.5),        
        ('droprate_l', 0.5),
        ('log_count_l', 0.1),
        ('prob_l2r', 0.1),
        ('log_count_l2r', 0.1),
        ('known_LR', 1.0),
        ('R_is_syllable', -0.1),
        ('log_length', 0.5)
])

ScoreTable = namedtuple('ScoreTable', list(default_profile))
Table = namedtuple('Table', 'L R begin end length, lr_prop lr_count cohesion_l droprate_l lcount')

class LREvaluator:
    def __init__(self, profile=None):
        self.profile = profile if profile else default_profile

    def evaluate(self, candidates, preference=None):
        scores = []
        for c in candidates:
            score = self._evaluate(
                              self.make_scoretable(c.L[0],
                              c.L[1],
                              c.R[0],
                              c.R[1],
                              c.cohesion_l,
                              c.droprate_l,
                              c.lcount,
                              c.lr_prop,
                              c.lr_count,
                              c.length
                         ))
            if preference:
                if c.L[1] and c.L[1] in preference:
                    score += preference.get(c.L[1], {}).get(c.L[0], 0)
                if c.R[1] and c.R[1] in preference:
                    score += preference.get(c.R[1], {}).get(c.R[0], 0)
            scores.append((c, score))
        return sorted(scores, key=lambda x:-x[-1])

    def make_scoretable(self, l, pos_l, r, pos_r, cohesion, droprate, lcount, lr_prop, lr_count, len_LR):
        return ScoreTable(cohesion,
                          droprate,
                          log(lcount+1),
                          lr_prop,
                          log(lr_count+1),
                          1 if (pos_l and pos_r) else 0,
                          1 if len(r) == 1 else 0,
                          log(len_LR)
                         )

    def _evaluate(self, scoretable):
        return sum(score * self.profile.get(field, 0) for field, score in scoretable._asdict().items())


class LRMaxScoreTagger:
    def __init__(self, domain_dictionary_folders=None, use_base_dictionary=True,
                 dictionary_word_mincount=3,
                 evaluator=None, sents=None, lrgraph=None, 
                 lrgraph_lmax=12, lrgraph_rmax=8,
                 base_tokenizer=None, preference=None, verbose=False
                ):
        
        self.dictionary = Dictionary(domain_dictionary_folders, use_base_dictionary, dictionary_word_mincount, verbose=verbose)
        self.evaluator = evaluator if evaluator else LREvaluator()
        self.preference = preference if preference else {}
        self.lrgraph = lrgraph if lrgraph else {}
        
        if (not self.lrgraph) and (sents):
            self.lrgraph = _build_lrgraph(sents, lrgraph_lmax, lrgraph_rmax)
            
        self.lrgraph_norm, self.lcount, self.cohesion_l, self.droprate_l\
            = self._initialize_scores(self.lrgraph)

        self.base_tokenizer = base_tokenizer if base_tokenizer else lambda x:x.split()
        if not base_tokenizer:
            try:
                self.base_tokenizer = MaxScoreTokenizer(scores=self.cohesion_l)
            except Exception as e:
                print('MaxScoreTokenizer(cohesion) exception: {}'.format(e))
        
    def _build_lrgraph(self, sents, lmax=12, rmax=8):
        from collections import Counter
        from collections import defaultdict
        eojeols = Counter((eojeol for sent in sents for eojeol in sent.split() if eojeol))
        lrgraph = defaultdict(lambda: defaultdict(int))
        for eojeol, count in eojeols.items():
            n = len(eojeol)
            for i in range(1, min(n, lmax)+1):
                (l, r) = (eojeol[:i], eojeol[i:])
                if len(r) > rmax:
                    continue
                lrgraph[l][r] += count
                
        return lrgraph
    
    def _initialize_scores(self, lrgraph):
        def to_counter(dd):
            return {k:sum(d.values()) for k,d in dd.items()}
        def to_normalized_graph(dd):
            normed = {}
            for k,d in dd.items():
                sum_ = sum(d.values())
                normed[k] = {k1:c/sum_ for k1,c in d.items()}
            return normed

        lrgraph_norm = to_normalized_graph(lrgraph)
        lcount = to_counter(lrgraph)
        cohesion_l = {w:pow(c/lcount[w[0]], 1/(len(w)-1)) for w, c in lcount.items() if len(w) > 1}
        droprate_l = {w:c/lcount[w[:-1]] for w, c in lcount.items() if len(w) > 1 and w[:-1] in lcount}
        
        return lrgraph_norm, lcount, cohesion_l, droprate_l
    
    def pos(self, sent, flatten=True, debug=False):
        sent_ = [self._pos(eojeol, debug) for eojeol in sent.split() if eojeol]
        if flatten:
            sent_ = [word for words in sent_ for word in words]
        return sent_

    def _pos(self, eojeol, debug=False):
        candidates = self._initialize(eojeol)
        scores = self._scoring(candidates)
        best = self._find_best(scores)
        if best:
            post = self._postprocessing(eojeol, best)
        else:
            post = self._base_tokenizing_subword(eojeol, 0)
            
        if not debug:
            post = [w for lr in post for w in lr[:2] if w[0]]
        return post
    
    def _initialize(self, t):
        candidates = self._initialize_L(t)
        candidates = self._initialize_LR(t, candidates)
        return candidates

    def _initialize_L(self, t):
        n = len(t)
        candidates = []
        for b in range(n):
            for e in range(b+2, min(n, b+self.dictionary._lmax)+1):
                l = t[b:e]
                l_pos = self.dictionary.pos_L(l)
                if not l_pos:
                    continue

                candidates.append([l,       # 0
                                   l_pos,       # 1
                                   b,      # 2                             
                                   e,      # 3
                                   e-b,   # 4
                                  ])

        candidates = self._remove_l_subsets(candidates)
        return sorted(candidates, key=lambda x:x[2])

    def _remove_l_subsets(self, candidates):
        candidates_ = []
        for pos in ['Noun', 'Verb', 'Adjective', 'Adverb', 'Exclamation']:
            # Sort by len_L
            sorted_ = sorted(filter(lambda x:x[1] == pos, candidates), key=lambda x:-x[4])
            while sorted_:
                candidates_.append(sorted_.pop(0))
                (b, e) = (candidates_[-1][2], candidates_[-1][3])
    #             removals = [i for i, c in enumerate(sorted_) if b < c[3] and e > c[2]] # Overlap
                removals = [i for i, c in enumerate(sorted_) if b <= c[2] and e >= c[3]] # Subset (Contain)
                for idx in reversed(removals):
                    del sorted_[idx]
        return candidates_

    def _initialize_LR(self, t, candidates, threshold_prop=0.001, threshold_count=2):
        n = len(t)
        expanded = []

        for (l, pos, b, e, len_l) in candidates:        
            for len_r in range(min(self.dictionary._rmax, n-e)+1):

                r = t[e:e+len_r]
                lr_prop = self.lrgraph_norm.get(l, {}).get(r, 0)
                lr_count = self.lrgraph.get(l, {}).get(r, 0)

                if (r) and ((lr_prop <= threshold_prop) or (lr_count <= threshold_count)):
                    continue

                expanded.append([(l, pos),
                                 (r, None if not r else self.dictionary.pos_R(r)),
                                 b,
                                 e,
                                 e + len_r,
                                 len_r,
                                 len_l + len_r,
                                 lr_prop,
                                 lr_count
                                ])

        expanded = self._remove_r_subsets(expanded)
        return sorted(expanded, key=lambda x:x[2])

    def _remove_r_subsets(self, expanded):
        expanded_ = []
        for pos in ['Josa', 'Verb', 'Adjective', None]:
            # Sory by len_R
            sorted_ = sorted(filter(lambda x:x[1][1] == pos, expanded), key=lambda x:-x[5])
            while sorted_:
                expanded_.append(sorted_.pop(0))
                (b, e) = (expanded_[-1][3], expanded_[-1][4])
    #             removals = [i for i, c in enumerate(sorted_) if b < c[3] and e > c[2]] # Overlap
                removals = [i for i, c in enumerate(sorted_) if b <= c[3] and e >= c[4]] # Subset (Contain)
                for idx in reversed(removals):
                    del sorted_[idx]
        expanded_ = [[L, R, p0, p2, len_LR, prop, count] for L, R, p0, p1, p2, len_R, len_LR, prop, count in expanded_]
        return expanded_
    
    def _scoring(self, candidates):
        candidates = [self._to_table(c) for c in candidates]
        scores = self.evaluator.evaluate(candidates, self.preference if self.preference else None)
        return scores

    def _to_table(self, c):
        return Table(c[0], c[1], c[2], c[3], c[4], c[5], c[6], 
                     self.cohesion_l.get(c[0][0], 0),
                     self.droprate_l.get(c[0][0], 0),
                     self.lcount.get(c[0][0], 0)
                    )
    
    def _find_best(self, scores):
        best = []
        sorted_ = sorted(scores, key=lambda x:-x[-1])
        while sorted_:
            best.append(sorted_.pop(0)[0])
            (b, e) = (best[-1][2], best[-1][3])
            removals = [i for i, (c, _) in enumerate(sorted_) if b < c[3] and e > c[2]] # Overlap
            for idx in reversed(removals):
                del sorted_[idx]
        return sorted(best, key=lambda x:x[2])
    
    def _postprocessing(self, t, words):
        n = len(t)
        adds = []
        if words and words[0][2] > 0:
            adds += self._add_first_subword(t, words)
        if words and words[-1][3] < n:
            adds += self._add_last_subword(t, words, n)
        adds += self._add_inter_subwords(t, words)
        post = [w for w in words] + [self._to_table(a) for a in adds]
        return sorted(post, key=lambda x:x[2])

    def _infer_subword_information(self, subword):
        pos = self.dictionary.pos_L(subword)
        prop = self.lrgraph_norm.get(subword, {}).get('', 0.0)
        count = self.lrgraph.get(subword, {}).get('', 0)    
        if not pos:
            pos = self.dictionary.pos_R(subword)
        return (pos, prop, count)

    def _add_inter_subwords(self, t, words):
        adds = []        
        for i, base in enumerate(words[:-1]):
            if base[3] == words[i+1][2]:
                continue

            b = base[3]
            e = words[i+1][2]
            subword = t[b:e]
            #(pos, prop, count) = self._infer_subword_information(subword)
            #adds.append([(subword, pos), ('', None), b, e, e-b, prop, count, 0.0])
            adds += self._base_tokenizing_subword(subword, b)
        return adds

    def _add_last_subword(self, t, words, n):
        b = words[-1][3]
        subword = t[b:]
        #(pos, prop, count) = self._infer_subword_information(subword)
        #return [[(subword, pos), ('', None), b, n, n-b, prop, count, 0.0]]
        return self._base_tokenizing_subword(subword, b)

    def _add_first_subword(self, t, words):    
        e = words[0][2]
        subword = t[0:e]
        #(pos, prop, count) = self._infer_subword_information(subword)
        #return [[(subword, pos), ('', None), 0, e, e, prop, count, 0.0]]
        return self._base_tokenizing_subword(subword, 0)
    
    def _base_tokenizing_subword(self, t, b):
        subwords = []
        _subwords = self.base_tokenizer.tokenize(t, flatten=False)
        if not _subwords:
            return []
        for w in _subwords[0]:
            (pos, prop, count) = self._infer_subword_information(w[0])
            subwords.append([(w[0], pos), ('', None), b+w[1], b+w[2], w[2]-w[1], prop, count, 0.0])
        return subwords
    
    def add_words_into_dictionary(self, words, tag):
        if not (tag in self.dictionary._pos):
            raise ValueError('{} does not exist base dictionary'.format(tag))
        self.dictionary.add_words(words, tag)
        
    def remove_words_from_dictionary(self, words, tag):
        if not (tag in self.dictionary._pos):
            raise ValueError('{} does not exist base dictionary'.format(tag))
        self.dictionary.remove_words(words, tag)
    
    def save_domain_dictionary(self, folder, head=None):
        self.dictionary.save_domain_dictionary(folder, head)
    
    def set_word_preferance(self, words, tag, preference=10):
        if type(words) == str:
            words = {words}
        preference_table = self.preference.get(tag, {})
        preference_table.update({word:preference for word in words})
        self.preference[tag] = preference_table
    
    def save_tagger(self, fname):
        raise NotImplemented


from ._normalizer import emoticon_normalize
from ._normalizer import repeat_normalize
from ._normalizer import only_hangle
from ._normalizer import only_hangle_number
from ._normalizer import only_text

__all__ = ['emoticon_normalize',
           'repeat_normalize',
           'only_hangle',
           'only_hangle_number',
           'only_text'
          ]


# -*- encoding:utf8 -*-
import sys
if sys.version_info <= (2,7):
    reload(sys)
    sys.setdefaultencoding('utf-8')
from pprint import pprint
import re
import numpy as np


class RegexTokenizer:
    
    def __init__(self):
        self._patterns = [
            ('number', re.compile(u'[-+]?\d*[\.]?[\d]+|[-+]?\d+', re.UNICODE)),
            ('korean', re.compile(u'[가-힣]+', re.UNICODE)),
            ('jaum', re.compile(u'[ㄱ-ㅎ]+', re.UNICODE)),
            ('moum', re.compile(u'[ㅏ-ㅣ]+', re.UNICODE)),
            ('english & latin', re.compile(u"[a-zA-ZÀ-ÿ]+[[`']?s]*|[a-zA-ZÀ-ÿ]+", re.UNICODE))
        ]
        
        self.doublewhite_pattern = re.compile('\s+')

    def __call__(self, s, debug=True, flatten=True):
        return self.tokenize(s, debug, flatten)

    def tokenize(self, s, debug=False, flatten=True):
        '''
        Usage
        
        s = "이거에서+3.12같은34숫자나-1.2like float해해 같은aÀÿfafAis`s-1찾아서3.1.2.1해ㅋㅋㅜㅠ봐 Bob`s job.1"
        tokenizer = RegularTokenizer()
        tokenizer.tokenize(s)

        [['이거에서', '+3.12', '같은', '34', '숫자나', '-1.2', 'like'],
         ['float', '해해'],
         ['같은', 'aÀÿfafAis`s', '-1', '찾아서', '3.1', '.2', '.1', '해', 'ㅋㅋ', 'ㅜㅠ', '봐'],
         ['Bob`s'],
         ['job', '.1']]
        '''
        tokens = [self._tokenize(t, debug) for t in s.split()]
        if flatten:
            tokens = [subtoken for token in tokens for subtoken in token if subtoken]
        return tokens
    
    def _tokenize(self, s, debug=False):
        for name, pattern in self._patterns:
            
            founds = pattern.findall(s)
            if not founds: 
                continue

            if debug:
                print('\n%s' % name)
                print(founds)
            
            found = founds.pop(0)
            len_found = len(found)
            
            s_ = ''
            b = 0
            for i, c in enumerate(s):
                
                if b > i: 
                    continue
                    
                if s[i:i+len_found] == found:
                    s_ += ' %s ' % s[i:i+len_found]
                    b = i + len_found
                    
                    if not founds:
                        s_ += s[b:]
                        break
                    else:
                        found = founds.pop(0)
                        len_found = len(found)
                    
                    continue
                s_ += c
            s = s_
            
        s = self.doublewhite_pattern.sub(' ', s).strip().split()
        # TODO: handle 3.1.2.1
        return s


class LTokenizer:
    
    def __init__(self, scores=None, default_score=0.0):
        self._scores = scores if scores else {}
        self._ds = default_score

    def __call__(self, sentence, tolerance=0.0, flatten=True, remove_r=False):
        return self.tokenize(sentence, tolerance, flatten, remove_r)

    def tokenize(self, sentence, tolerance=0.0, flatten=True, remove_r=False):
        
        def token_to_lr(token, tolerance=0.0):
            length = len(token)
            if length <= 2: return (token, '')
            candidates = [(token[:e], token[e:]) for e in range(2, length + 1)]
            candidates = [(self._scores.get(t[0], self._ds), t[0], t[1]) for t in candidates]
            if tolerance > 0:
                max_score = max([c[0] for c in candidates])
                candidates = [c for c in candidates if (max_score - c[0]) <= tolerance]
                best = sorted(candidates, key=lambda x:len(x[1]), reverse=True)[0]
            else:
                best = sorted(candidates, key=lambda x:(x[0], len(x[1])), reverse=True)[0]
            return (best[1], best[2])

        tokens = [token_to_lr(token, tolerance) for token in sentence.split()]
        
        if remove_r:
            tokens = [token[0] for token in tokens]
        
        if (flatten) and (remove_r == False):
            tokens = [subtoken for token in tokens for subtoken in token if subtoken]
        
        return tokens
    

class MaxScoreTokenizer:
    
    def __init__(self, scores=None, max_length=10, default_score=0.0):
        self._scores = scores if scores else {}
        self._max_length = max_length
        self._ds = default_score

    def __call__(self, sentence, flatten=True):
        return self.tokenize(sentence, flatten)

    def tokenize(self, sentence, flatten=True):
        tokens = [self._recursive_tokenize(token) for token in sentence.split()]
        if flatten:
            tokens = [subtoken[0] for token in tokens for subtoken in token]
        return tokens

    def _recursive_tokenize(self, token, range_l=0, debug=False):
        
        length = len(token)
        if length <= 2:
            return [(token, 0, length, self._ds, length)]

        if range_l == 0:
            range_l = min(self._max_length, length)

        scores = self._initialize(token, range_l, length)
        if debug:
            pprint(scores)
        
        result = self._find(scores)
        
        adds = self._add_inter_subtokens(token, result)
        
        if result[-1][2] != length:
            adds += self._add_last_subtoken(token, result)
            
        if result[0][1] != 0:
            adds += self._add_first_subtoken(token, result)
            
        return sorted(result + adds, key=lambda x:x[1])

    def _initialize(self, token, range_l, length):
        scores = []
        for b in range(0, length - 1):
            for r in range(2, range_l + 1):
                e = b + r
                
                if e > length: 
                    continue
                
                subtoken = token[b:e]
                score = self._scores.get(subtoken, self._ds)
                scores.append((subtoken, b, e, score, r))
                
        return sorted(scores, key=lambda x:(-x[3], -x[4], x[1]))

    def _find(self, scores):
        result = []
        num_iter = 0
        
        while scores:
            word, b, e, score, r = scores.pop(0)
            result.append((word, b, e, score, r))

            if not scores:
                break

            removals = []
            for i, (_1, b_, e_, _2, _3) in enumerate(scores):
                if (b_ < e and b < e_) or (b_ < e and e_ > b):
                    removals.append(i)

            for i in reversed(removals):
                del scores[i]

            num_iter += 1
            if num_iter > 100: break

        return sorted(result, key=lambda x:x[1])
    
    def _add_inter_subtokens(self, token, result):
        adds = []        
        for i, base in enumerate(result[:-1]):
            if base[2] == result[i+1][1]:
                continue
            
            b = base[2]
            e = result[i+1][1]
            subtoken = token[b:e]
            adds.append((subtoken, b, e, self._ds, e - b))
        
        return adds

    def _add_first_subtoken(self, token, result):
        e = result[0][1]
        subtoken = token[0:e]
        score = self._scores.get(subtoken, self._ds)
        return [(subtoken, 0, e, score, e)]
    
    def _add_last_subtoken(self, token, result):
        b = result[-1][2]
        subtoken = token[b:]
        score = self._scores.get(subtoken, self._ds)
        return [(subtoken, b, len(token), score, len(subtoken))]

class MaxLRScoreTokenizer:
    def __init__(self, Dl=None, Dr=None,
                 preference_l=None, preference_r=None,
                 lrgraph=None, tokenizer_builder=None,
                 max_lscore_difference=0.3, max_lscore_diffratio=0.5, # Expansion L
                 ensurable_score_l=0.5, ensurable_score_lr_diff=0.3   # R overlap L
                ):

        # Normalize L-R graph to prob graph
        def norm(rdict):
            sum_ = sum(rdict.values())
            return {r:c/sum_ for r,c in rdict.items()}
        self.lrgraph = lrgraph if lrgraph else {}
        self.lrgraph_norm = {l:norm(rdict) for l,rdict in self.lrgraph.items()}

        # Expanding dictionary from lrgraph
        #self.Dl, self.Dr = tokenizer_builder(self.lrgraph) if tokenizer_builder else LRTokenizerBuilder()(self.lrgraph)
        self.Dl, self.Dr = tokenizer_builder(self.lrgraph) if tokenizer_builder else ({}, {})
        
        # Dictionary type check
        if not Dl: Dl = {}
        if not Dr: Dr = {}

        if not type(Dl) == dict:
            Dl = {l:1.0 for l in Dl}
        self.Dl.update(Dl)
        if not type(Dr) == dict:
            Dr = {r:1.0 for r in Dr}
        self.Dr.update(Dr)

        # Add preference words into dictionary
        self.Pl = preference_l if preference_l else {}
        self.Pr = preference_r if preference_r else {}

        for l in self.Pl:
            if not (l in self.Dl):
                self.Dl[l] = 1.0
        for r in self.Pr:
            if not (r in self.Dr):
                self.Dr[r] = 1.0

        self.lmax = max((len(w) for w in self.Dl)) if self.Dl else 0
        self.rmax = max((len(w) for w in self.Dr)) if self.Dr else 0
        self.base_tokenizer = MaxScoreTokenizer(scores=self.Dr)
        
        self.max_lscore_difference = max_lscore_difference
        self.max_lscore_diffratio = max_lscore_diffratio
        self.ensurable_score_l = ensurable_score_l
        self.ensurable_score_lr_diff = ensurable_score_lr_diff

    def __call__(self, sent, debug=True, flatten=True):
        return self.tokenize(sent, debug, flatten)

    def tokenize(self, sent, debug=False, flatten=True):
        sent_ = [self._tokenize(t, debug) for t in sent.split() if t]
        if flatten:
            sent_ = [word for words in sent_ for word in words]
        return sent_

    def _tokenize(self, t, debug=False):
        candidates = self._initialize(t)
        candidates_ = self._remove_l_subset(candidates)
        scores = self._score(candidates_)
        best = self._find_best(scores)
        
        if best:
            post = self._postprocessing(t, best)
        else:
            post = self._base_tokenizing_subword(t, 0)

        if not debug:
            post = [[(p[0], 'L'), (p[1], 'R')] for p in post]
            post = [w for p in post for w in p if w[0]]
        return post
    
    def _initialize(self, t):
        candidates = self._initialize_L(t)
        candidates = self._initialize_LR(t, candidates)
        return candidates

    def _initialize_L(self, t):
        n = len(t)
        candidates = []
        for b in range(n):
            for e in range(b+1, min(n, b+self.lmax)+1):
                l = t[b:e]
                if not (l in self.Dl):
                    continue
                candidates.append([l,     # 0
                                   b,     # 1                  
                                   e,     # 2
                                   e-b    # 3
                                  ])
        return candidates

    def _initialize_LR(self, t, candidates):
        n = len(t)
        expanded = []
        for (l, b, e, len_l) in candidates:
            for len_r in range(min(self.rmax, n-e)+1):
                if len_l == 1 and len_r == 0:
                    continue
                r = t[e:e+len_r]
                if r and not (r in self.Dr):
                    continue
                expanded.append([l,
                                 r,
                                 b,
                                 e,
                                 e + len_r,
                                 len_l,
                                 len_r,
                                 len_l + len_r,
                                ])
        return sorted(expanded, key=lambda x:x[4])
    
    def _remove_l_subset(self, candidates):    
        for c in candidates:
            c.append(self.Dl.get(c[0], 0))
            c.append(self.Dr.get(c[1], 0))
        candidates = sorted(candidates, key=lambda x:-x[-2])

        candidates_ = []
        while candidates:
            best = candidates.pop(0)
            b, e, lscore = best[2], best[3], best[-2]

            exist_longer = False
            for c in candidates:
                if c[2] > b or c[3] < e or not (c[2] < b or c[3] > e):
                    continue
                if ((lscore - c[-2]) < self.max_lscore_difference) or \
                    ((self.ensurable_score_l * 0.5 < lscore) and \
                        ((lscore+1e-5) / (c[-2]+1e-5) < self.max_lscore_diffratio)):
                    exist_longer = True
                    break

            if not exist_longer:
                candidates_.append(best)

        return candidates_

    def _score(self, candidates):
        from collections import defaultdict
        # With checking R is overlapped next L
        begin_to_words = defaultdict(lambda: [])
        for c in candidates:
            begin_to_words[c[2]].append(c)
        begin_to_words = dict(begin_to_words)

        scored = []

        candidates = sorted(candidates, key=lambda x:(-x[-2], -x[-1], x[2], -x[5]))
        while candidates:
            c = candidates.pop(0)
            l, r, p0, p1, p2, len_l, len_r, len_lr, score_l, score_r = c

            # Check whether R is overlapped next L
            if len_r:
                overlappped = False
                for b in range(p1, p2):
                    if overlappped:
                        break
                    for word in begin_to_words.get(b, []):
                        score_diff = word[-2] + self.Pl.get(word[0], 0) - score_r
                        if (self.ensurable_score_l <= word[-2]) or (score_diff > self.ensurable_score_lr_diff):
                            overlappped = True
                            break
                if overlappped:
                    continue

            total_score = (score_l * 2 if not r else score_l + score_r) + self.Pl.get(l, 0) + self.Pr.get(r, 0)
            c.append(total_score)
            scored.append(c)
        return scored

    def _find_best(self, scores):
        best = []
        sorted_ = sorted(scores, key=lambda x:-x[-1])
        while sorted_:
            best.append(sorted_.pop(0))
            (b, e) = (best[-1][2], best[-1][4])
            removals = [i for i, c in enumerate(sorted_) if b < c[4] and e > c[2]] # Overlap
            for idx in reversed(removals):
                del sorted_[idx]
        return sorted(best, key=lambda x:x[2])

    def _postprocessing(self, t, words):
        n = len(t)
        adds = []
        if words and words[0][2] > 0:
            adds += self._add_first_subword(t, words)
        if words and words[-1][3] < n:
            adds += self._add_last_subword(t, words, n)
        adds += self._add_inter_subwords(t, words)
        post = [w for w in words] + adds
        return sorted(post, key=lambda x:x[2])

    def _add_inter_subwords(self, t, words):
        adds = []        
        for i, base in enumerate(words[:-1]):
            if base[4] == words[i+1][2]:
                continue
            b = base[4]
            e = words[i+1][2]
            subword = t[b:e]
            adds += self._base_tokenizing_subword(subword, b)
        return adds

    def _add_last_subword(self, t, words, n):
        b = words[-1][3]
        subword = t[b:]
        return self._base_tokenizing_subword(subword, b)

    def _add_first_subword(self, t, words):    
        e = words[0][2]
        subword = t[0:e]
        return self._base_tokenizing_subword(subword, 0)

    def _base_tokenizing_subword(self, t, b):
        words = self.base_tokenizer.tokenize(t)
        words_ = []
        b_ = 0
        for w in words:
            n = len(w)
            # TODO: 여기를 바꿔야해
            if w in self.Dr:
                words_.append(['', w, b+b_, b+b_, b+n, 0, n, n, 0, self.base_tokenizer.scores[w]])
            else:
                words_.append([w, '', b+b_, b+n, b+n, n, 0, n, 0, 0])
        return words_


# -*- encoding:utf8 -*-

import sys
if sys.version_info <= (2,7):
    reload(sys)
    sys.setdefaultencoding('utf-8')
import re
from soynlp.hangle import decompose

doublespace_pattern = re.compile('\s+')
repeatchars_pattern = re.compile('(\w)\\1{3,}')
hangle_filter = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣]')
hangle_number_filter = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣0-9]')
text_filter = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9\.\?\!\"\'-]')


def repeat_normalize(sent, num_repeats=2):
    if num_repeats > 0:
        sent = repeatchars_pattern.sub('\\1' * num_repeats, sent)
    sent = doublespace_pattern.sub(' ', sent)
    return sent.strip()

def emoticon_normalize(sent, num_repeats=2):
    if not sent:
        return sent

    # Pattern matching ㅋ쿠ㅜ
    def pattern(idx):
        # Jaum: 0, Moum: 1, Complete: 2, else -1
        if 12593 <= idx <= 12622:
            return 0
        elif 12623 <= idx <= 12643:
            return 1
        elif 44032 <= idx <= 55203:
            return 2
        else:
            return -1

    idxs = [pattern(ord(c)) for c in sent]
    sent_ = []
    for i, (idx, c) in enumerate(zip(idxs[:-1], sent)):
        if i > 0 and (idxs[i-1] == 0 and idx == 2 and idxs[i+1] == 1):
            cho, jung, jong = decompose(sent[i])
            if (cho == sent[i-1]) and (jung == sent[i+1]) and (jong == ' '):
                sent_.append(cho)
                sent_.append(jung)
            else:
                sent_.append(c)
        else:
            sent_.append(c)
    sent_.append(sent[-1])
    return repeat_normalize(''.join(sent_), num_repeats)

def only_hangle(sent):
    return doublespace_pattern.sub(' ',hangle_filter.sub(' ', sent)).strip()

def only_hangle_number(sent):
    return doublespace_pattern.sub(' ',hangle_number_filter.sub(' ', sent)).strip()

def only_text(sent):
    return doublespace_pattern.sub(' ',text_filter.sub(' ', sent)).strip()


from soynlp.utils import get_process_memory
from collections import defaultdict
from scipy.sparse import csr_matrix

def sent_to_word_contexts_matrix(sents, windows=3, min_tf=10,
        tokenizer=lambda x:x.split(), dynamic_weight=False, verbose=True):

    """
    :param dynamic_weight : Use dynamic weight if True.
        co-occurrence weight = [1, (w-1)/w, (w-2)/w, ... 1/w]
    """

    if verbose:
        print('Create (word, contexts) matrix')

    vocab2idx, idx2vocab = _scanning_vocabulary(
        sents, min_tf, tokenizer, verbose)

    word2contexts = _word_context(
        sents, windows, tokenizer, dynamic_weight, verbose, vocab2idx)

    x = _encode_as_matrix(word2contexts, vocab2idx, verbose)

    if verbose:
        print('  - done')
    return x, idx2vocab

def _scanning_vocabulary(sents, min_tf, tokenizer, verbose):

    # counting word frequency, first
    word_counter = defaultdict(int)

    for i_sent, sent in enumerate(sents):

        if verbose and i_sent % 1000 == 0:
            _print_status('  - counting word frequency', i_sent)

        words = tokenizer(sent)
        for word in words:
            word_counter[word] += 1

    if verbose:
        _print_status('  - counting word frequency', i_sent, new_line=True)

    # filtering with min_tf    
    vocab2idx = {word for word, count in word_counter.items() if count >= min_tf}
    vocab2idx = {word:idx for idx, word in enumerate(
        sorted(vocab2idx, key=lambda w:-word_counter[w]))}
    idx2vocab = [word for word, _ in sorted(vocab2idx.items(), key=lambda w:w[1])]
    del word_counter

    return vocab2idx, idx2vocab

def _print_status(message, i_sent, new_line=False):
    print('\r{} from {} sents, mem={} Gb'.format(
            message, i_sent, '%.3f' % get_process_memory()),
        flush=True, end='\n' if new_line else ''
    )

def _word_context(sents, windows, tokenizer, dynamic_weight, verbose, vocab2idx):

    # scanning (word, context) pairs
    word2contexts = defaultdict(lambda: defaultdict(int))

    if dynamic_weight:
        weight = [(windows-i)/windows for i in range(windows)]
    else:
        weight = [1] * windows

    for i_sent, sent in enumerate(sents):

        if verbose and i_sent % 1000 == 0:
            _print_status('  - scanning (word, context) pairs', i_sent)

        words = tokenizer(sent)
        if not words:
            continue

        n = len(words)

        for i, word in enumerate(words):
            if not (word in vocab2idx):
                continue

            # left_contexts
            for w in range(windows):
                j = i - (w + 1)
                if j < 0 or not (words[j] in vocab2idx):
                    continue
                word2contexts[word][words[j]] += weight[w]

            # right_contexts
            for w in range(windows):
                j = i + w + 1
                if j >= n or not (words[j] in vocab2idx):
                    continue
                word2contexts[word][words[j]] += weight[w]

    if verbose:
        _print_status('  - scanning (word, context) pairs', i_sent, new_line=True)

    return word2contexts

def _encode_as_matrix(word2contexts, vocab2idx, verbose):

    rows = []
    cols = []
    data = []
    for word, contexts in word2contexts.items():
        word_idx = vocab2idx[word]
        for context, cooccurrence in contexts.items():
            context_idx = vocab2idx[context]
            rows.append(word_idx)
            cols.append(context_idx)
            data.append(cooccurrence)
    x = csr_matrix((data, (rows, cols)))

    if verbose:
        print('  - (word, context) matrix was constructed. shape = {}{}'.format(
            x.shape, ' '*20))

    return x


# -*- encoding:utf8 -*-

from soynlp.hangle import compose, decompose

class Lemmatizer:
    def __init__(self, stems, endings, predefined=None):
        self._stems = stems
        self._endings = endings
        self._initialize()
        if predefined:
            self._predefined.update(predefined)

    def _initialize(self):
        self._predefined = {'불어':('붇다', '불다'),
                            '그래':('그렇다',)
                           }

    def lemmatize(self, word, check_only_stem=False):
        candidates = set()
        for i in range(1, len(word)+1):
            l, r = word[:i], word[i:]
            for stem, ending in _lemma_candidate(l, r, self._predefined):
                if stem in self._stems:
                    if check_only_stem:
                        candidates.add((stem, ending))
                    elif ending in self._endings:
                        candidates.add((stem, ending))
        return candidates

    def candidates(self, word):
        candidates = set()
        for i in range(1, len(word) + 1):
            l = word[:i]
            r = word[i:]
            candidates.update(self._lemma_candidate(l, r, self._predefined))
        return candidates

def _lemma_candidate(l, r, predefined=None):
    def add_lemma(stem, ending):
        candidates.add((stem, ending))

    candidates = {(l, r)}

    l_last = decompose(l[-1])
    l_last_ = compose(l_last[0], l_last[1], ' ')
    l_front = l[:-1]
    r_first = decompose(r[0]) if r else ('', '', '')
    r_first_ = compose(r_first[0], r_first[1], ' ') if r else ' '
    r_end = r[1:]

    # ㄷ 불규칙 활용: 깨달 + 아 -> 깨닫 + 아
    if l_last[2] == 'ㄹ' and r_first[0] == 'ㅇ':
        l_stem = l_front + compose(l_last[0], l_last[1], 'ㄷ')
        add_lemma(l_stem, r)

    # 르 불규칙 활용: 굴 + 러 -> 구르 + 어
    if (l_last[2] == 'ㄹ') and (r_first_ == '러' or r_first_ == '라'):
        l_stem = l_front + compose(l_last[0], l_last[1], ' ') + '르'
        r_canon = compose('ㅇ', r_first[1], r_first[2]) + r_end
        add_lemma(l_stem, r_canon)

    # ㅂ 불규칙 활용: 더러 + 워서 -> 더럽 + 어서
    if (l_last[2] == ' ') and (r_first_ == '워' or r_first_ == '와'):
        l_stem = l_front + compose(l_last[0], l_last[1], 'ㅂ')
        r_canon = compose('ㅇ', 'ㅏ' if r_first_ == '와' else 'ㅓ', r_first[2]) + r_end
        add_lemma(l_stem, r_canon)

    # 어미의 첫글자가 종성일 경우 (-ㄴ, -ㄹ, -ㅂ, -ㅅ)
    # 입 + 니다 -> 이 + ㅂ니다
    if l_last[2] == 'ㄴ' or l_last[2] == 'ㄹ' or l_last[2] == 'ㅂ' or l_last[2] == 'ㅆ':
        l_stem = l_front + compose(l_last[0], l_last[1], ' ')
        r_canon = l_last[2] + r
        add_lemma(l_stem, r_canon)

    # ㅅ 불규칙 활용: 부 + 어 -> 붓 + 어
    # exception : 벗 + 어 -> 벗어
    if (l_last[2] == ' ' and l[-1] != '벗') and (r_first[0] == 'ㅇ'):
        l_stem = l_front + compose(l_last[0], l_last[1], 'ㅅ')
        add_lemma(l_stem, r)

    # 우 불규칙 활용: 똥퍼 + '' -> 똥푸 + 어
    if l_last_ == '퍼':
        l_stem = l_front + '푸'
        r_canon = compose('ㅇ', l_last[1], l_last[2]) + r
        add_lemma(l_stem, r_canon)

    # 우 불규칙 활용: 줬 + 어 -> 주 + 었어
    if l_last[1] == 'ㅝ':
        l_stem = l_front + compose(l_last[0], 'ㅜ', ' ')
        r_canon = compose('ㅇ', 'ㅓ', l_last[2]) + r
        add_lemma(l_stem, r_canon)

    # 오 불규칙 활용: 왔 + 어 -> 오 + 았어
    if l_last[1] == 'ㅘ':
        l_stem = l_front + compose(l_last[0], 'ㅗ', ' ')
        r_canon = compose('ㅇ', 'ㅏ', l_last[2]) + r
        add_lemma(l_stem, r_canon)

    # ㅡ 탈락 불규칙 활용: 꺼 + '' -> 끄 + 어 / 텄 + 어 -> 트 + 었어
    if (l_last[1] == 'ㅓ' or l_last[1] == 'ㅏ'):
        l_stem = l_front + compose(l_last[0], 'ㅡ', ' ')
        r_canon = compose('ㅇ', l_last[1], l_last[2]) + r
        add_lemma(l_stem, r_canon)

    # 거라, 너라 불규칙 활용
    # '-거라/-너라'를 어미로 취급하면 규칙 활용
    # if (l[-1] == '가') and (r and (r[0] == '라' or r[:2] == '거라')):
    #    # TODO

    # 러 불규칙 활용: 이르 + 러 -> 이르다
    # if (r_first[0] == 'ㄹ' and r_first[1] == 'ㅓ'):
    #     if self.is_stem(l):
    #         # TODO

    # 여 불규칙 활용
    # 하 + 였다 -> 하 + 았다 -> 하다: '였다'를 어미로 취급하면 규칙 활용

    # 여 불규칙 활용 (2)
    # 했 + 다 -> 하 + 았다 / 해 + 라니깐 -> 하 + 아라니깐 / 했 + 었다 -> 하 + 았었다
    if l_last[0] == 'ㅎ' and l_last[1] == 'ㅐ':
        l_stem = l_front + '하'
        r_canon = compose('ㅇ', 'ㅏ', l_last[2]) + r
        add_lemma(l_stem, r_canon)

    # ㅎ (탈락) 불규칙 활용
    if (l_last[2] == ' ' or l_last[2] == 'ㄴ' or l_last[2] == 'ㄹ' or l_last[2] == 'ㅂ' or l_last[2] == 'ㅆ'):
        # 파라 + 면 -> 파랗 + 면
        if (l_last[1] == 'ㅏ' or l_last[1] == 'ㅓ'):
            l_stem = l_front + compose(l_last[0], l_last[1], 'ㅎ')
            r_canon = r if l_last[2] == ' ' else l_last[2] + r
            add_lemma(l_stem, r_canon)
        # ㅎ (축약) 불규칙 할용
        # 시퍼렜 + 다 -> 시퍼렇 + 었다, 파랬 + 다 -> 파랗 + 았다
        if (l_last[1] == 'ㅐ') or (l_last[1] == 'ㅔ'):
            # exception : 그렇 + 아 -> 그래
            if len(l) >= 2 and l[-2] == '그' and l_last[0] == 'ㄹ':
                l_stem = l_front + '렇'
            else:
                l_stem = l_front + compose(l_last[0], 'ㅓ' if l_last[1] == 'ㅔ' else 'ㅏ', 'ㅎ')
            r_canon = compose('ㅇ', 'ㅓ' if l_last[1] == 'ㅔ' else 'ㅏ', l_last[2]) + r
            add_lemma(l_stem, r_canon)

    # 이었 -> 였 규칙활용
    # 좋아졌 + 어 -> 좋아지 + 었어, 좋아졋 + 던 -> 좋아지 + 었던, 좋아져 + 서 -> 좋아지 + 어서
    # 였 + 어 -> 이 + 었어
    # 종성 ㅆ 을 ㅅ 으로 쓰는 경우도 고려
    if ((l_last[2] == 'ㅆ' or l_last[2] == 'ㅅ' or l_last[2] == ' ') and
        (l_last[1] == 'ㅕ') or (l_last[1] == 'ㅓ')):

        # except: -었 -> 이 + 었 (x) // -였-> 이 + 었 (o) // -졌 -> 지 + 었 (o) // -젔 -> 지 + 었
        if ((l_last[0] == 'ㅇ') and (l_last[1] == 'ㅕ')) or not (l_last[0] == 'ㅇ'):
            l_stem = l_front + compose(l_last[0], 'ㅣ', ' ')
            r_canon = compose('ㅇ', 'ㅓ', l_last[2])+ r
            add_lemma(l_stem, r_canon)

    ## Pre-defined set
    if predefined and (l, r) in predefined:
        for stem in predefined[(l, r)]:
            candidates.add(stem)

    return candidates


# -*- encoding:utf8 -*-

from ._template import LR

class BaseTagger:
    def __init__(self, generator, evaluator, postprocessor=None):
        self.evaluator = evaluator
        self.generator = generator
        self.dictionary = generator.dictionary
        self.postprocessor = postprocessor

    def tag(self, sentence, flatten=True, debug=False):
        raise NotImplementedError

class SimpleTagger(BaseTagger):
    def tag(self, sentence, flatten=True, debug=False):
        sent_ = []
        debug_ = []
        eojeols = sentence.split()
        
        for eojeol in eojeols:
            candidates = self.generator.generate(eojeol)
            best = self.evaluator.select_best(candidates)
            
            if self.postprocessor:
                postprocessed = self.postprocessor.postprocess(eojeol, best)
            else:
                postprocessed = best
            
            # Flatten in a token
            postprocessed_ = []
            for word in postprocessed:
                if word.l:
                    postprocessed_.append((word.l, word.l_tag))
                if word.r:
                    postprocessed_.append((word.r, word.r_tag))
            
            sent_.append(postprocessed_)
            
            if debug:
                scored_candidates = [(c, self.evaluator.evaluate(c)) for c in candidates]
                scored_candidates = sorted(scored_candidates, key=lambda x:(x[0].b, x[1]))
                debug_.append(scored_candidates)
        
        # Flatten in a sentence
        if flatten:
            sent_ = [word for words in sent_ for word in words]

        if not debug:
            return sent_
        else:
            return sent_, debug_

class BasePostprocessor:
    def postprocess(self, token, best_wordstream):
        return best_wordstream

class UnknowLRPostprocessor(BasePostprocessor):
    def postprocess(self, token, words):
        n = len(token)
        adds = []
        if words and words[0].b > 0:
            adds.append(self._add_first_subword(token, words))
        if words and words[-1].e < n:
            adds.append(self._add_last_subword(token, words, n))
        adds += self._add_inter_subwords(token, words)
        post = words + adds
        return sorted(post, key=lambda x:x.b)

    def _add_last_subword(self, token, words, n):
        b = words[-1].e
        subword = token[b:]
        return LR(subword, None, '', None, b, n, n)

    def _add_first_subword(self, token, words):    
        e = words[0].b
        subword = token[0:e]
        return LR(subword, None, '', None, 0, e, e)
    
    def _add_inter_subwords(self, token, words):
        adds = []        
        for i, base in enumerate(words[:-1]):
            if base.e == words[i+1].b:
                continue
            b = base.e
            e = words[i+1].b
            subword = token[b:e]
            adds.append(LR(subword, None, '', None, b, e, e))
        return adds


from ._distance import levenshtein
from ._distance import jamo_levenshtein
from ._distance import cosine_distance
from ._distance import jaccard_distance
from ._hangle import normalize
from ._hangle import decompose
from ._hangle import compose
from ._hangle import character_is_korean
from ._hangle import character_is_complete_korean
from ._hangle import character_is_jaum
from ._hangle import character_is_moum
from ._hangle import to_base
from ._hangle import character_is_number
from ._hangle import character_is_english
from ._hangle import ConvolutionHangleEncoder